from glob import glob
import time
import os
from datetime import datetime

import torch
import torch.backends.cudnn as cudnn
from torch.optim.lr_scheduler import ReduceLROnPlateau
from fastai import *
from fastai.core import *
from fastai.torch_core import *
from fastai.basic_data import *
from fastai.basic_train import LearnerCallback, Learner
from catalyst.data.sampler import DistributedSamplerWrapper, BalanceClassSampler
import torch_xla.distributed.xla_multiprocessing as xmp
import torch_xla.distributed.parallel_loader as pl
import torch_xla.core.xla_model as xm
import torch_xla.utils.utils as xu
import torch_xla.distributed.data_parallel as dp

from kaggle_runner import logger
from kaggle_runner.kernels.fastai_kernel import FastAIKernel
from kaggle_runner.metrics.meters import AverageMeter, RocAucMeter
from kaggle_runner.data_providers import provider
from kaggle_runner.logs import metric_get_log
from kaggle_runner.losses import MixedLoss
from kaggle_runner.metrics.meters import Meter
from kaggle_runner.optimizers import RAdam


class TPUDistributed(LearnerCallback):
    def __init__(self, learn: Learner, debug=True):
        super().__init__(learn)

        self.debug = debug

        if debug:
            self.device = xm.xla_device(devkind='TPU')
            logger.debug("TPUDistributed in DEBUG mode")
            #self.device = xm.xla_device(devkind='CPU')
        else:
            self.device = xm.xla_device(devkind='TPU')
        logger.debug("%s used for xla_device for TPUDistributed" % self.device)

    def on_train_begin(self, **kwargs: Any) -> None:
        self.learn.model = self.learn.model.to(self.device)

        pg = self.learn.opt.opt.param_groups
        pg0pl = pg[0]['params']  # pg0pl[0] is a Parameter
        pg1pl = pg[1]['params']  # pg0pl[0] is a Parameter

        #logger.debug("grad info: %s", raw_opt)
        logger.debug(f"on_train_begin pg0 lr: {pg[0]['lr']}")
        logger.debug(f"on_train_begin pg1 lr: {pg[1]['lr']}")

        if self.debug:
            self.learn.opt.lr = self.learn.opt.lr*xm.xrt_world_size()
            #pg[0]['lr'] *= xm.xrt_world_size() # will do it twice...
            #pg[1]['lr'] *= xm.xrt_world_size()
            logger.debug("opt info: %s\n type: %s",
                         self.learn.opt, type(self.learn.opt))
        else:
            self.learn.opt.lr = self.learn.opt.lr*xm.xrt_world_size()

        logger.debug("%s used for xla_device, to device done" % self.device)

        shuffle = self.data.train_dl.init_kwargs['shuffle'] if hasattr(
            self.data.train_dl, 'init_kwargs') else True
        self.old_sampler_train_dl, self.data.train_dl, self.train_sampler = _change_dl(
            self.k,
            self.data.train_dl, shuffle)

        if hasattr(self.data, 'valid_dl') and self.data.valid_dl is not None:
            self.old_sampler_valid_dl, self.data.valid_dl, self.valid_sampler = _change_dl_val(
                self.k, self.data.valid_dl, shuffle)

    def on_epoch_begin(self, **kwargs: Any) -> None:
        logger.debug("Epoch begins on device %s" % self.device)

        self.old_train_dl = self.data.train_dl
        self.learn.data.train_dl = pl.ParallelLoader(

            self.old_train_dl, [self.device]).per_device_loader(self.device)
        self.learn.data.train_dl.dataset = None  # self.old_train_dl.dataset

        if hasattr(self.data, 'valid_dl') and self.data.valid_dl is not None:
            self.old_valid_dl = self.learn.data.valid_dl
            self.learn.data.valid_dl = pl.ParallelLoader(
                self.old_valid_dl, [self.device]).per_device_loader(self.device)

            self.learn.data.valid_dl.dataset = self.old_valid_dl.dataset
            self.learn.data.valid_dl.dl = self.learn.data.valid_dl._loader._loader

    def on_backward_end(self, **kwargs: Any) -> None:
        # copied from https://github.com/tmabraham/fastai_tpu/blob/8b73018cf705da1a73d9be1f105a8e886051a90c/fastai_v1/tpu_distributed_fastai.py, and needed a fix
        xm.optimizer_step(self.learn.opt.opt, barrier=True)
        #may_debug(True)

        return {'skip_step': True}

    def on_epoch_end(self, **kwargs: Any) -> None:
        self.learn.data.train_dl = self.old_train_dl
        self.learn.data.valid_dl = self.old_valid_dl

    def on_train_end(self, **kwargs: Any) -> None:
        self.learn.data.train_dl = self.old_sampler_train_dl
        self.learn.data.valid_dl = self.old_sampler_valid_dl


class TPUFitter:
    def __init__(self, model, device, config):

        if not os.path.exists('node_submissions'):
            os.makedirs('node_submissions')

        self.config = config
        self.epoch = 0
        self.log_path = 'log.txt'

        self.model = model
        self.device = device

        param_optimizer = list(self.model.named_parameters())
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        optimizer_grouped_parameters = [
            {'params': [p for n, p in param_optimizer if not any(
                nd in n for nd in no_decay)], 'weight_decay': 0.001},
            {'params': [p for n, p in param_optimizer if any(
                nd in n for nd in no_decay)], 'weight_decay': 0.0}
        ]

        self.optimizer = AdamW(optimizer_grouped_parameters,
                               lr=config.lr*xm.xrt_world_size())
        self.scheduler = config.SchedulerClass(
            self.optimizer, **config.scheduler_params)

        self.criterion = config.criterion
        xm.master_print(f'Fitter prepared. Device is {self.device}')

    def fit(self, train_loader, validation_loader):
        for e in range(self.config.n_epochs):
            if self.config.verbose:
                lr = self.optimizer.param_groups[0]['lr']
                timestamp = datetime.utcnow().isoformat()
                self.log(f'\n{timestamp}\nLR: {lr}')

            t = time.time()
            para_loader = pl.ParallelLoader(train_loader, [self.device])
            losses, final_scores = self.train_one_epoch(
                para_loader.per_device_loader(self.device))

            self.log(
                f'[RESULT]: Train. Epoch: {self.epoch}, loss: {losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, time: {(time.time() - t):.5f}')

            t = time.time()
            para_loader = pl.ParallelLoader(validation_loader, [self.device])
            losses, final_scores = self.validation(
                para_loader.per_device_loader(self.device))

            self.log(
                f'[RESULT]: Validation. Epoch: {self.epoch}, loss: {losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, time: {(time.time() - t):.5f}')

            if self.config.validation_scheduler:
                self.scheduler.step(metrics=final_scores.mc_avg)

            self.epoch += 1

    def run_tuning_and_inference(self, test_loader, validation_tune_loader):
        for e in range(2):
            self.optimizer.param_groups[0]['lr'] = self.config.lr * \
                xm.xrt_world_size()
            para_loader = pl.ParallelLoader(
                validation_tune_loader, [self.device])
            losses, final_scores = self.train_one_epoch(
                para_loader.per_device_loader(self.device))
            para_loader = pl.ParallelLoader(test_loader, [self.device])
            self.run_inference(para_loader.per_device_loader(self.device))

    def validation(self, val_loader):
        self.model.eval()
        losses = AverageMeter()
        final_scores = RocAucMeter()

        t = time.time()

        for step, (inputs_masks, targets) in enumerate(val_loader):
            inputs = inputs_masks[0]
            attention_masks = inputs_masks[1]

            if self.config.verbose:
                if step % self.config.verbose_step == 0:
                    xm.master_print(
                        f'Valid Step {step}, loss: ' +
                        f'{losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, ' +
                        f'time: {(time.time() - t):.5f}'
                    )
            with torch.no_grad():
                inputs = inputs.to(self.device, dtype=torch.long)
                attention_masks = attention_masks.to(
                    self.device, dtype=torch.long)
                targets = targets.to(self.device, dtype=torch.float)

                outputs = self.model(inputs, attention_masks)
                loss = self.criterion(outputs, targets)

                batch_size = inputs.size(0)

                final_scores.update(targets, outputs)
                losses.update(loss.detach().item(), batch_size)

        return losses, final_scores

    def train_one_epoch(self, train_loader):
        self.model.train()

        losses = AverageMeter()
        final_scores = RocAucMeter()
        t = time.time()

        for step, (inputs_masks, targets) in enumerate(train_loader):
            inputs = inputs_masks[0]
            attention_masks = inputs_masks[1]
            batch_size = inputs.size(0)

            if self.config.verbose:
                if step % self.config.verbose_step == 0:
                    self.log(
                        f'Train Step {step}, loss: ' +
                        f"{losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, " +
                        f'time: {(time.time() - t):.5f}'
                    )

            inputs = inputs.to(self.device, dtype=torch.long)
            attention_masks = attention_masks.to(self.device, dtype=torch.long)
            targets = targets.to(self.device, dtype=torch.float)

            self.optimizer.zero_grad()

            outputs = self.model(inputs, attention_masks)
            loss = self.criterion(outputs, targets)

            final_scores.update(targets, outputs)

            losses.update(loss.detach().item(), batch_size)

            loss.backward()
            #logger.info("step: %d, loss: %f", step, loss)

            xm.optimizer_step(self.optimizer)

            if self.config.step_scheduler:
                self.scheduler.step()

        self.model.eval()
        self.save('last-checkpoint.bin')

        return losses, final_scores

    def run_inference(self, test_loader):
        self.model.eval()
        result = {'id': [], 'toxic': []}
        t = time.time()

        for step, (inputs_masks, ids) in enumerate(test_loader):
            inputs = inputs_masks[0]
            attention_masks = inputs_masks[1]

            if self.config.verbose:
                if step % self.config.verbose_step == 0:
                    xm.master_print(
                        f'Prediction Step {step}, time: {(time.time() - t):.5f}')

            with torch.no_grad():
                inputs = inputs.to(self.device, dtype=torch.long)
                attention_masks = attention_masks.to(
                    self.device, dtype=torch.long)
                outputs = self.model(inputs, attention_masks)
                toxics = nn.functional.softmax(
                    outputs, dim=1).data.cpu().numpy()[:, 1]

            result['id'].extend(ids.cpu().numpy())
            result['toxic'].extend(toxics)

        result = pd.DataFrame(result)
        node_count = len(glob('node_submissions/*.csv'))
        result.to_csv(
            f'node_submissions/submission_{node_count}_{datetime.utcnow().microsecond}_{random.random()}.csv', index=False)

    def save(self, path):
        xm.save(self.model.state_dict(), path)

    def log(self, message):
        if self.config.verbose:
            xm.master_print(message)
        with open(self.log_path, 'a+') as logger:
            xm.master_print(f'{message}', logger)



from __future__ import print_function

import collections
import functools
import os
import random
import types  # for bound new forward function for RoIHeads

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data
import torchvision
import torchvision.models.detection.roi_heads as roi_heads
import torchvision.transforms.functional as TF
from PIL import Image, ImageFile
from torchvision import transforms
from torchvision.models.detection import _utils as det_utils
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
from torchvision.models.detection.roi_heads import (fastrcnn_loss,
                                                    keypointrcnn_inference,
                                                    keypointrcnn_loss,
                                                    maskrcnn_inference,
                                                    maskrcnn_loss)
from torchvision.ops import boxes as box_ops
from torchvision.ops import roi_align
from tqdm import tqdm

import kaggle_runner.kernels.KernelRunningState
import kaggle_runner.logs
import kaggle_runner.optimizers
from kaggle_runner.kernels.kernel import KaggleKernel
from kaggle_runner.utils import kernel_utils

ImageFile.LOAD_TRUNCATED_IMAGES = True


# Plot inline
# %matplotlib inline


class PS_torch(KaggleKernel):
    def __init__(self, *args, **kargs):
        super(PS_torch, self).__init__(*args, **kargs)

        self.model_ft = None
        self.data_loader_dev = None
        self.lr_scheduler = None

        # data
        self._stat_dataset = None
        self.img_mean = [0.46877811, 0.46877811, 0.46877811]
        self.img_std = [0.24535184, 0.24535184, 0.24535184]
        # default value: mean:[0.46877811 0.46877811 0.46877811],
        # std:[0.24535184 0.24535184 0.24535184]

        # for debugging thing
        self.metric_logger = kaggle_runner.logs.MetricLogger(delimiter="  ")
        self.metric_logger.add_meter(
            "lr", kaggle_runner.logs.SmoothedValue(window_size=1, fmt="{value:.6f}")
        )
        self.metric_logger.add_meter(
            "loss", kaggle_runner.logs.SmoothedValue(window_size=160, fmt="{avg:.6f}")
        )
        self.metric_logger.add_meter(
            "loss_mask",
            kaggle_runner.logs.SmoothedValue(window_size=160, fmt="{avg:.6f}"),
        )
        self.DATA_PATH_BASE = "../input/siimacr-pneumothorax-segmentation-data-128"

    def dump_state(
        self, exec_flag=False, force=True
    ):  # only dataloader ... others cannot dumped
        kernel_utils.logger.debug(f"state {self._stage}")
        if exec_flag:
            kernel_utils.logger.debug(f"dumping state {self._stage}")
            self_data = vars(self)

            if self.model_ft is not None:
                torch.save(self.model_ft.state_dict(), "cv_model.pth")
            names_to_exclude = {
                "model_ft",
                "optimizer",
                "lr_scheduler",
                "metric_logger",
            }

            data_to_save = {
                k: v for k, v in self_data.items() if k not in names_to_exclude
            }

            kernel_utils.dump_obj(
                data_to_save, f"run_state_{self._stage}.pkl", force=force
            )

            # print(self_data)
            # for k, v in self_data.items():
            #    utils.dump_obj(v, f'run_state_{k}_{self._stage}.pkl')

    def load_state_data_only(self, stage, file_name="run_state.pkl"):
        if stage is not None:
            file_name = f"run_state_{stage}.pkl"
        kernel_utils.logger.debug(f"restore from {file_name}")
        self_data = kernel_utils.get_obj_or_dump(filename=file_name)

        self._stage = self_data["_stage"]

        # self.model_ft = self_data['model_ft']
        self.num_epochs = self_data["num_epochs"]
        # self.optimizer = self_data['optimizer']
        self.data_loader = self_data["data_loader"]
        self.data_loader_dev = self_data["data_loader_dev"]
        self.device = self_data["device"]
        # self.lr_scheduler = self_data['lr_scheduler']

    @staticmethod
    def eval_model_loss(
        model, data_loader, device, metric_logger, print_freq, mode="train"
    ):
        # metric_logger = utils.MetricLogger(delimiter="  ")
        header = "Evaluation"
        losses_summed = 0.0
        cnt = 0
        metric_logger.clear()

        # model.eval() # will output box, seg, scores
        if mode == "train":
            model.train()
        else:
            model.eval()  # will output box, seg, scores

        with torch.no_grad():
            for images, targets in metric_logger.log_every(
                data_loader, print_freq, header
            ):
                images = list(image.to(device) for image in images)
                targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

                if mode != "train":
                    predictions, _ = model(images, targets)
                    continue
                loss_dict = model(images, targets)
                loss_dict.pop("loss_rpn_box_reg")
                # diferrent model use different names
                loss_dict.pop("loss_box_reg")

                # reduce losses over all GPUs for logging purposes
                loss_dict_reduced = kernel_utils.reduce_dict(loss_dict)
                losses_reduced = sum(loss for loss in loss_dict_reduced.values())
                losses_summed += losses_reduced.detach().cpu().numpy()
                cnt += 1

                metric_logger.update(loss=losses_reduced, **loss_dict_reduced)

        metric_logger.clear()
        if mode != "train":
            return predictions
        return losses_summed / cnt

    @staticmethod
    def eval_model(model, data_loader, device, metric_logger, print_freq):
        # metric_logger = utils.MetricLogger(delimiter="  ")
        header = "Evaluation"
        losses_reduced = float("inf")
        metric_logger.clear()

        # model.eval() # will output box, seg, scores
        with torch.no_grad():
            for images, targets in metric_logger.log_every(
                data_loader, print_freq, header
            ):
                images = list(image.to(device) for image in images)
                targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

                loss_dict = model(images, targets)

                losses = sum(loss for loss in loss_dict.values())

                # reduce losses over all GPUs for logging purposes
                loss_dict_reduced = kernel_utils.reduce_dict(loss_dict)
                losses_reduced = sum(loss for loss in loss_dict_reduced.values())

                metric_logger.update(loss=losses_reduced, **loss_dict_reduced)

        metric_logger.clear()
        return losses_reduced.cpu().numpy()

    @staticmethod
    def train_one_epoch(
        model,
        optimizer,
        data_loader,
        device,
        epoch,
        metric_logger,
        print_freq,
        mq_logger=None,
    ):
        model.train()
        # metric_logger = utils.MetricLogger(delimiter="  ")
        # metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1,
        # fmt='{value:.6f}'))
        header = "Epoch: [{}]".format(epoch)

        metric_logger.clear()

        losses_summed = 0.0
        cnt = 0

        warm_up_lr_scheduler = None
        if epoch == 0:
            warmup_factor = 1.0 / 1000
            warmup_iters = min(1000, len(data_loader) - 1)

            warm_up_lr_scheduler = kernel_utils.warmup_lr_scheduler(
                optimizer, warmup_iters, warmup_factor
            )

        for images, targets in metric_logger.log_every(data_loader, print_freq, header):
            images = list(image.to(device) for image in images)
            targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

            loss_dict = model(images, targets)
            # it just pop.. and we do not train rpn anyway!!!!
            loss_dict.pop("loss_rpn_box_reg")
            # diferrent model use different names
            loss_dict.pop("loss_box_reg")

            losses = sum(loss for loss in loss_dict.values())

            # reduce losses over all GPUs for logging purposes
            loss_dict_reduced = kernel_utils.reduce_dict(loss_dict)
            losses_reduced = sum(loss for loss in loss_dict_reduced.values())

            optimizer.zero_grad()
            losses.backward()
            optimizer.step()

            losses_summed += losses_reduced.detach().cpu().numpy()
            cnt += 1

            if warm_up_lr_scheduler is not None:  # only for epoch 0, warm up
                warm_up_lr_scheduler.step()

            if mq_logger is not None:  # issue is it runs off... So NAN
                mq_logger.debug(f"losses summed is {losses_summed}, cnt is {cnt}")
                print(
                    f"losses summed is {losses_summed}, cnt is {cnt}, loss_dict_reduced is {loss_dict_reduced}"
                )
            metric_logger.update(loss=losses_reduced, **loss_dict_reduced)
            metric_logger.update(lr=optimizer.param_groups[0]["lr"])

        return losses_summed / cnt

    @staticmethod
    def _collate_fn_for_data_loader(x):
        return tuple(zip(*x))

    def prepare_train_dev_data(self):
        df = pd.read_csv(self.DATA_PATH_BASE + "/train-rle.csv")
        try:
            if self._debug_less_data:
                df = pd.read_csv(self.DATA_PATH_BASE + "/train-rle.csv")[:100]
        except Exception:
            pass

        imgdir = self.DATA_PATH_BASE + "/train/"

        if self.img_mean is None or self.img_std is None:
            self._stat_dataset = SIIMDataset_split_df(
                df, imgdir, no_aug=True
            )  # for mean calculation

        if self.submit_run:
            df_train = df
        else:
            train_mask = np.zeros((len(df),), dtype=np.bool)
            np.random.seed(2019)
            chosen = np.random.choice(len(df), int(0.8 * len(df)), replace=False)
            train_mask[chosen] = True
            val_mask = np.invert(train_mask)

            df_train = df[train_mask]
            df_dev = df[val_mask]

            dataset_dev = SIIMDataset_split_df(df_dev, imgdir, no_aug=True)
            self.data_loader_dev = torch.utils.data.DataLoader(
                dataset_dev,
                batch_size=4,
                shuffle=False,
                num_workers=4,  # 4: 08:19, 8: 08:40
                collate_fn=self._collate_fn_for_data_loader,
            )

        dataset_train = SIIMDataset_split_df(df_train, imgdir)
        # print(dataset_train[2019][1]['area'])  # only for debug
        # print(dataset_dev[19][1]['area'])

        kernel_utils.logger.debug("torch train dataloader initializing")
        self.data_loader = torch.utils.data.DataLoader(
            dataset_train,
            batch_size=4,
            shuffle=True,
            num_workers=4,  # 4: 08:19, 8: 08:40
            collate_fn=self._collate_fn_for_data_loader,
        )

    def after_prepare_data_hook(self):
        if self.img_mean is not None and self.img_std is not None:
            return

        stat_data_loader = torch.utils.data.DataLoader(
            self._stat_dataset,
            batch_size=8,
            shuffle=False,
            num_workers=4,
            collate_fn=self._collate_fn_for_data_loader,
        )
        self.img_mean, self.img_std = kernel_utils.online_mean_and_sd(
            stat_data_loader, data_map=lambda x: x[0]
        )
        self.metric_logger.print_and_log_to_file(
            f"mean:{self.img_mean}, std:{self.img_std}"
        )
        del stat_data_loader

    def build_and_set_model(self):
        # create mask rcnn model
        num_classes = 2
        self.device = torch.device(
            "cpu"
        )  # TODO check if cuda is supported, or we just use cpu

        # more details at https://pytorch.org/tutorials/intermediate/torchvision_tutorial.html

        # finetuning

        # load a model pre-trained on COCO, num_classes=91, cannot change.... as the pretrained model won't load
        self.model_ft = torchvision.models.detection.maskrcnn_resnet50_fpn(
            pretrained=True, image_mean=self.img_mean, image_std=self.img_std
        )
        # FL = FocalLoss(gamma=2, alpha=0.75)  # Version 5, gamma=2, alpha=0.75, 0.8034...
        # FL = FocalLoss(gamma=1, alpha=0.75, magnifier=3)  # + early stop,2 stop at 7, (data split) version 7 0.8026
        # FL = FocalLoss(gamma=1, alpha=0.5, magnifier=3)  # version 8 0.8031, 6 epoch
        # FL = FocalLoss(gamma=0.5, alpha=0.5, magnifier=1)  # command line submission, 4 epochs cv+aug
        FL = FocalLoss(
            gamma=0.5, alpha=0.5, magnifier=1
        )  # changed lr decay 2/0.15 + patience=3, do not use focal loss...
        FL_wrapped = functools.partial(maskrcnn_loss_focal, focal_loss_func=FL)
        # FL_wrapped = None  # changed lr decay 2/0.15, do not use focal loss... 0.8025

        RoIHeads_loss_customized.set_customized_loss(
            self.model_ft.roi_heads, maskrcnn_loss_customized=FL_wrapped
        )
        RoIHeads_loss_customized.update_forward_func(self.model_ft.roi_heads)

        # get number of input features for the classifier
        in_features = self.model_ft.roi_heads.box_predictor.cls_score.in_features
        # replace the pre-trained head with a new one
        self.model_ft.roi_heads.box_predictor = FastRCNNPredictor(
            in_features, num_classes
        )

        # change mask prediction head, only predict background and pneu... part
        in_features_mask = self.model_ft.roi_heads.mask_predictor.conv5_mask.in_channels
        hidden_layer = 256
        self.model_ft.roi_heads.mask_predictor = MaskRCNNPredictor(
            in_features_mask, hidden_layer, num_classes
        )

        # GPU
        # my_trace()  # test about to
        self.model_ft.to(self.device)
        #  self.logger.debug(f"model info:\n{self.model_ft}")

        # for param in self.model_ft.parameters():
        #    param.requires_grad = True

        params = [p for p in self.model_ft.parameters() if p.requires_grad]

        start_learning_rate = 0.001

        try:
            if self._debug_continue_training:  # monkey patch
                start_learning_rate = 0.00001
        except Exception:
            pass
        self.optimizer = torch.optim.SGD(
            params, lr=start_learning_rate, momentum=0.9, weight_decay=0.0005
        )
        self.lr_scheduler = torch.optim.lr_scheduler.StepLR(
            self.optimizer,
            step_size=4,
            # after changed to 3, 0.8042 (improved from 0.8033)
            # step_size 4, with little aug, 0.8037 (Version 11)
            gamma=0.1,
        )

    def train_model(self):
        patience = 3
        if self.submit_run:
            patience = 0
        es = kaggle_runner.optimizers.EarlyStopping(
            patience=patience
        )  # the first time it become worse, if patience set to 1

        for epoch in range(self.num_epochs):
            train_loss = self.train_one_epoch(
                self.model_ft,
                self.optimizer,
                self.data_loader,
                self.device,
                epoch,
                self.metric_logger,
                print_freq=10,
                mq_logger=self.logger,
            )
            self.metric_logger.print_and_log_to_file(
                f"train_loss (averaged) is {train_loss}"
            )
            self.lr_scheduler.step()  # change learning rate

            if not self.submit_run:
                metric = self.eval_model_loss(
                    self.model_ft,
                    self.data_loader_dev,
                    self.device,
                    self.metric_logger,
                    print_freq=10,
                )
                self.metric_logger.print_and_log_to_file(
                    f"\nmetric (averaged) is {metric}\n"
                )
                if es.step(metric):
                    self.print_log(
                        f"{epoch+1} epochs run and early stop, with patience {patience}"
                    )
                    break

    def print_log(self, s):
        self.metric_logger.print_and_log_to_file(s)

    def save_model(self):
        torch.save(self.model_ft.state_dict(), "cv_model.pth")

    def load_model_weight_continue_train(self):
        self.build_and_set_model()
        assert os.path.exists("cv_model.pth")
        self.model_ft.load_state_dict(torch.load("cv_model.pth"))
        self.model_ft.eval()

        patience = 3
        if self.submit_run:
            patience = 0
        es = kaggle_runner.optimizers.EarlyStopping(
            patience=patience
        )  # the first time it become worse, if patience set to 1
        for epoch in range(8, 13):
            train_loss = self.train_one_epoch(
                self.model_ft,
                self.optimizer,
                self.data_loader,
                self.device,
                epoch,
                self.metric_logger,
                print_freq=100,
            )
            print(f"train_loss (averaged) is {train_loss}")
            self.lr_scheduler.step()  # change learning rate

            if not self.submit_run:
                metric = self.eval_model_loss(
                    self.model_ft,
                    self.data_loader_dev,
                    self.device,
                    self.metric_logger,
                    print_freq=100,
                )
                print(f"metric (averaged) is {metric}")
                if es.step(metric):
                    print(
                        f"{epoch+1} epochs run and early stop, with patience {patience}"
                    )
                    break

    def load_model_weight(self):
        self.build_and_set_model()
        self.model_ft.load_state_dict(torch.load("cv_model.pth"))
        self.model_ft.eval()

    def pre_test(self):
        # change model to prediction mode
        for param in self.model_ft.parameters():
            param.requires_grad = False

        self.model_ft.eval()

    def predict_rle_from_acts_with_threshold(self, acts, threshold):
        return self.predict_rle_from_acts(
            acts, self.data_loader_dev, self.metric_logger, 100, threshold
        )

    @staticmethod
    def predict_rle_from_acts(
        acts, data_loader, metric_logger, print_freq, threshold=0.5
    ):
        header = "DEV prediction test"

        sublist = []
        counter = 0

        data_loader_b1 = torch.utils.data.DataLoader(
            data_loader.dataset,
            batch_size=1,
            shuffle=False,
            num_workers=0,  # 4: 08:19, 8: 08:40
            collate_fn=PS_torch._collate_fn_for_data_loader,
        )

        idx = 0
        mask_add_cnt = []
        for image, target in metric_logger.log_every(
            data_loader_b1, print_freq, header
        ):
            width, height = 1024, 1024

            result = acts[idx]
            image_id = target[0]["image_id"].cpu().numpy()
            idx += 1
            if len(result["masks"]) > 0:
                counter += 1
                mask_added = 0
                for ppx in range(len(result["masks"])):
                    if result["scores"][ppx] >= threshold:
                        mask_added += 1
                        # res = transforms.ToPILImage()(result["masks"][ppx].permute(1, 2, 0).cpu().numpy())
                        res = transforms.ToPILImage()(
                            result["masks"][ppx].transpose((1, 2, 0))
                        )
                        res = np.asarray(
                            res.resize((width, height), resample=Image.BILINEAR)
                        )
                        res = (res[:, :] * 255.0 > 127).astype(np.uint8).T
                        rle = kernel_utils.mask_to_rle(res, width, height)
                        sublist.append([image_id, rle])
                if mask_added == 0:
                    rle = " -1"
                    sublist.append([image_id, rle])
                mask_add_cnt.append(mask_added)
            else:
                rle = " -1"
                sublist.append([image_id, rle])
                mask_add_cnt.append(0)  # no mask
            if idx % 100 == 0:
                mask_stat = np.array(mask_add_cnt)
                metric_logger.update(
                    **{
                        "mask0": ((mask_stat == 0).sum() / idx),
                        "mask1": ((mask_stat == 1).sum() / idx),
                        "mask2": ((mask_stat == 2).sum() / idx),
                        "mask>2": ((mask_stat > 2).sum() / idx),
                    }
                )

        mask_stat = np.array(mask_add_cnt)
        assert len(mask_stat) == idx

        metric_logger.print_and_log_to_file(
            f"image cnt: {idx}, image predicted mask cnt: {counter}, "
            f"mask 0,1,2,3+ {(mask_stat==0).sum()/idx}, "
            f"{(mask_stat==1).sum()/idx}, {(mask_stat==2).sum()/idx}, {(mask_stat>2).sum()/idx}"
        )

        return sublist, mask_stat

    @staticmethod
    def predict_rle(model, device, data_loader, metric_logger, print_freq):
        header = "DEV prediction test"

        sublist = []
        counter = 0
        # changed from 0.25 to 0.5... need to check the data and analyze...(test on dev set)
        threshold = 0.55
        assert data_loader.batch_size == 1

        for images, targets in metric_logger.log_every(data_loader, print_freq, header):
            width, height = images.size
            image_id = targets["image_id"]

            images = list(image.to(device) for image in images)
            targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

            result = model(images, targets)[0]

            if len(result["masks"]) > 0:
                counter += 1
                mask_added = 0
                for ppx in range(len(result["masks"])):
                    if result["scores"][ppx] >= threshold:
                        mask_added += 1
                        res = transforms.ToPILImage()(
                            result["masks"][ppx].permute(1, 2, 0).cpu().numpy()
                        )
                        res = np.asarray(
                            res.resize((width, height), resample=Image.BILINEAR)
                        )
                        res = (res[:, :] * 255.0 > 127).astype(np.uint8).T
                        rle = kernel_utils.mask_to_rle(res, width, height)
                        sublist.append([image_id, rle])
                if mask_added == 0:
                    rle = " -1"
                    sublist.append([image_id, rle])
            else:
                rle = " -1"
                sublist.append([image_id, rle])

        return sublist

    def predict_on_test(self):
        sample_df = pd.read_csv(
            os.path.join(self.DATA_PATH_BASE, "sample_submission.csv")
        )

        # this part was taken from @raddar's kernel: https://www.kaggle.com/raddar/better-sample-submission
        masks_ = sample_df.groupby("ImageId")["ImageId"].count().reset_index(name="N")
        masks_ = masks_.loc[masks_.N > 1].ImageId.values
        ###
        sample_df = sample_df.drop_duplicates("ImageId", keep="last").reset_index(
            drop=True
        )

        tt = transforms.ToTensor()
        sublist = []
        counter = 0
        # changed from 0.25 to 0.5... need to check the data and analyze...(test on dev set)
        threshold = 0.55
        for index, row in tqdm(sample_df.iterrows(), total=len(sample_df)):
            image_id = row["ImageId"]
            if image_id in masks_:
                img_path = os.path.join(
                    self.DATA_PATH_BASE + "/test/" + image_id + ".png"
                )

                img = Image.open(img_path).convert("RGB")
                width, height = img.size
                img = img.resize((1024, 1024), resample=Image.BILINEAR)
                img = tt(img)
                result = self.model_ft([img.to(self.device)])[0]
                if len(result["masks"]) > 0:
                    counter += 1
                    mask_added = 0
                    for ppx in range(len(result["masks"])):
                        if result["scores"][ppx] >= threshold:
                            mask_added += 1
                            res = transforms.ToPILImage()(
                                result["masks"][ppx].permute(1, 2, 0).cpu().numpy()
                            )
                            res = np.asarray(
                                res.resize((width, height), resample=Image.BILINEAR)
                            )
                            res = (res[:, :] * 255.0 > 127).astype(np.uint8).T
                            rle = kernel_utils.mask_to_rle(res, width, height)
                            sublist.append([image_id, rle])
                    if mask_added == 0:
                        rle = " -1"
                        sublist.append([image_id, rle])
                else:
                    rle = " -1"
                    sublist.append([image_id, rle])
            else:
                rle = " -1"
                sublist.append([image_id, rle])

        submission_df = pd.DataFrame(sublist, columns=sample_df.columns.values)
        submission_df.to_csv("submission.csv", index=False)
        print(counter)

    def _build_show_model_detail(self):
        self.run(
            end_stage=kaggle_runner.kernels.KernelRunningState.KernelRunningState.PREPARE_DATA_DONE
        )
        self.build_and_set_model()
        # self.pre_test()
        # summary(self.model_ft, (1,1024,1024))

    def check_predict_details(self):
        # Thanks https://discuss.pytorch.org/t/how-can-l-load-my-best-model-as-a-feature-extractor-evaluator/17254/6
        assert self.model_ft is not None
        activations = kernel_utils.get_obj_or_dump("dev_output_results.pkl")
        self.analyzer = TorchModelAnalyzer(self)
        analyzer = self.analyzer

        if activations is not None:
            self.analyzer.activation = activations
        else:
            analyzer.register_forward_hook(
                self.model_ft.roi_heads, analyzer.get_output_saved("roi_heads")
            )
            # for mask_head output, size torch.Size([40, 256, 14, 14]),
            # (mask_fcn4): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
            # (relu4): ReLU(inplace)
            # for mask_predictor, torch.Size([46, 2, 28, 28])  (in evaluate mode)
            # for roi_heads, output.shape, tuple length 2, ([{'boxes', 'labels', 'scores':torch.Size([7]), 'masks':torch.Size([7, 1, 28, 28])},T,T,T],[ PLACE_FOR_LOSSES ])
            # for roi_heads, in training mode, output different number of test result, according to threshold thing

            self.eval_model_loss(
                self.model_ft,
                self.data_loader_dev,
                self.device,
                self.metric_logger,
                print_freq=150,
            )
            kernel_utils.dump_obj(
                analyzer.activation, "dev_output_results.pkl", force=True
            )

        roi_acts = []
        for acts in analyzer.activation["roi_heads"]:
            roi_acts += acts[0]
        self.analyzer.test_out_threshold(roi_acts)
        # my_trace()


class SIIMDataset(torch.utils.data.Dataset):
    def __init__(self, df_path, img_dir):
        self.df = pd.read_csv(df_path)
        self.height = 1024
        self.width = 1024
        self.image_dir = img_dir
        self.image_info = collections.defaultdict(dict)
        self.to_tensor_transformer = transforms.ToTensor()

        self.load_image_info()

    def load_image_info(self):
        counter = 0
        for index, row in tqdm(self.df.iterrows(), total=len(self.df)):
            image_id = row["ImageId"]
            image_path = os.path.join(self.image_dir, image_id)
            if (
                os.path.exists(image_path + ".png")
                and row[" EncodedPixels"].strip() != "-1"
            ):
                self.image_info[counter]["image_id"] = image_id
                self.image_info[counter]["image_path"] = image_path
                self.image_info[counter]["annotations"] = row[" EncodedPixels"].strip()
                counter += 1

    def _test_(self, idx):
        self.__getitem__(idx)

    def __getitem__(self, idx):
        img_path = self.image_info[idx]["image_path"]
        img = Image.open(img_path + ".png").convert(
            "RGB"
        )  # here it is converted to 3 channels
        width, height = img.size
        img = img.resize((self.width, self.height), resample=Image.BILINEAR)
        info = self.image_info[idx]

        mask = kernel_utils.rle2mask(info["annotations"], width, height)

        if mask is None:
            raise ValueError("mask is None!!!!!!!!")

        mask = Image.fromarray(mask.T)
        mask = mask.resize((self.width, self.height), resample=Image.BILINEAR)
        img, mask = self.functional_transforms(
            img, mask
        )  # pay attention need after mask.T
        mask = np.expand_dims(mask, axis=0)

        if sum(mask.flatten()) == 0:
            xmax = 0
            xmin = 0
            ymax = 0
            ymin = 0
        else:
            pos = np.where(np.array(mask)[0, :, :])
            xmin = np.min(pos[1])
            xmax = np.max(pos[1])
            ymin = np.min(pos[0])
            ymax = np.max(pos[0])

        boxes = torch.as_tensor([[xmin, ymin, xmax, ymax]], dtype=torch.float32)
        labels = torch.ones((1,), dtype=torch.int64)
        masks = torch.as_tensor(mask, dtype=torch.uint8)

        image_id = torch.tensor([idx])
        area = (boxes[:, 3] - boxes[:, 1]) * (boxes[:, 2] - boxes[:, 0])
        iscrowd = torch.zeros((1,), dtype=torch.int64)

        target = {}
        target["boxes"] = boxes
        target["labels"] = labels
        target["masks"] = masks
        target["image_id"] = image_id
        target["area"] = area
        target["iscrowd"] = iscrowd

        img = self.to_tensor_transformer(img)
        return img, target

    def functional_transforms(self, img, mask):
        # transforms.compose([
        #    transforms.ColorJitter(brightness=0.3, contrast=0.3),
        #    transforms.RandomAffine(degrees=15, translate=(0.1, 0.2)),
        #    transforms.RandomHorizontalFlip(p=0.5),
        #    transforms.RandomResizedCrop((self.width, self.height), scale=(0.2, 1.2), ratio=(0.9,1.1111))
        # ])
        pos_cnt = np.array(mask).sum()
        image = img
        segmentation = mask

        if random.random() > 0.5:
            while True:
                angle = random.uniform(-10, 10)
                translate_x = int(random.uniform(-0.05, 0.05) * self.width)
                translate_y = int(random.uniform(-0.1, 0.1) * self.height)
                scale = random.uniform(0.8, 1.2)

                image = TF.affine(
                    img, angle, (translate_x, translate_y), scale, shear=0
                )
                segmentation = TF.affine(
                    mask, angle, (translate_x, translate_y), scale, shear=0
                )
                if pos_cnt > 0:
                    trans_pos_cnt = np.array(segmentation).sum()
                    if trans_pos_cnt / (pos_cnt * scale) < 0.95:
                        continue  # not move the mask out

                # no more translation
                # angle = random.randint(-15, 15)
                # image = TF.rotate(img, angle)
                # segmentation = TF.rotate(mask, angle)
                # more transforms ...
                # bright_f = random.uniform(-0.3, 0.3)
                # image = TF.adjust_brightness(image, bright_f)
                ##segmentation = TF.adjust_brightness(segmentation, bright_f)

                # contrast_f = random.uniform(-0.3, 0.3)
                # image = TF.adjust_contrast(image, bright_f)
                ##segmentation = TF.adjust_contrast(segmentation, bright_f)
                # image = TF.resize(image, (self.height, self.width))
                # segmentation = TF.resize(segmentation, (self.height, self.width))  # the network can accept different size

                gamma = random.uniform(0.75, 1.3333)
                image = TF.adjust_gamma(image, gamma)
                break
        if random.random() > 0.5:
            image = TF.hflip(image)
            segmentation = TF.hflip(segmentation)

        return image, segmentation

    def __len__(self):
        return len(self.image_info)


class SIIMDataset_split_df(SIIMDataset):
    def __init__(self, df, img_dir, no_aug=False):
        # super(SIIMDataset_split_df, self).__init__()
        self.df = df
        self.height = 1024
        self.width = 1024
        self.image_dir = img_dir
        self.image_info = collections.defaultdict(dict)

        self.no_aug = no_aug

        self.to_tensor_transformer = transforms.ToTensor()
        self.load_image_info()

    def functional_transforms(self, img, mask):
        if self.no_aug:
            return img, mask
        else:
            return super(SIIMDataset_split_df, self).functional_transforms(img, mask)


# modified based https://github.com/DingKe/pytorch_workplace/blob/master/focalloss/loss.py
class FocalLoss(nn.Module):
    def __init__(self, gamma=0, alpha=0.5, eps=1e-7, magnifier=1.0, from_logits=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.eps = eps
        self.alpha = alpha
        self.magnifier = magnifier
        assert self.alpha >= 0
        assert self.alpha <= 1
        assert self.magnifier > 0
        self.from_logits = from_logits

    def forward(self, input, target):
        if self.from_logits:
            input = torch.sigmoid(input)

        y = target
        not_y = 1 - target

        y_hat = input
        not_y_hat = 1 - input

        y_hat = y_hat.clamp(self.eps, 1.0 - self.eps)
        not_y_hat = not_y_hat.clamp(self.eps, 1.0 - self.eps)

        loss = (
            -1 * self.alpha * not_y_hat ** self.gamma * y * torch.log(y_hat)
        )  # cross entropy
        loss += (
            -1 * (1 - self.alpha) * y_hat ** self.gamma * not_y * torch.log(not_y_hat)
        )
        loss *= self.magnifier

        return loss.mean()


class RoIHeads_loss_customized(roi_heads.RoIHeads):
    def __init__(
        self,
        box_roi_pool,
        box_head,
        box_predictor,
        # Faster R-CNN training
        fg_iou_thresh,
        bg_iou_thresh,
        batch_size_per_image,
        positive_fraction,
        bbox_reg_weights,
        # Faster R-CNN inference
        score_thresh,
        nms_thresh,
        detections_per_img,
        # Mask
        mask_roi_pool=None,
        mask_head=None,
        mask_predictor=None,
        keypoint_roi_pool=None,
        keypoint_head=None,
        keypoint_predictor=None,
        maskrcnn_loss_customized=None,
        fastrcnn_loss_customized=None,
        keypointrcnn_loss_customized=None,
    ):
        super(RoIHeads_loss_customized, self).__init__()

        self.box_similarity = box_ops.box_iou
        # assign ground-truth boxes for each proposal
        self.proposal_matcher = det_utils.Matcher(
            fg_iou_thresh, bg_iou_thresh, allow_low_quality_matches=False
        )

        self.fg_bg_sampler = det_utils.BalancedPositiveNegativeSampler(
            batch_size_per_image, positive_fraction
        )

        if bbox_reg_weights is None:
            bbox_reg_weights = (10.0, 10.0, 5.0, 5.0)
        self.box_coder = det_utils.BoxCoder(bbox_reg_weights)

        self.box_roi_pool = box_roi_pool
        self.box_head = box_head
        self.box_predictor = box_predictor

        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img

        self.mask_roi_pool = mask_roi_pool
        self.mask_head = mask_head
        self.mask_predictor = mask_predictor

        self.keypoint_roi_pool = keypoint_roi_pool
        self.keypoint_head = keypoint_head
        self.keypoint_predictor = keypoint_predictor

        self.maskrcnn_loss_customized = maskrcnn_loss_customized
        self.fastrcnn_loss_customized = fastrcnn_loss_customized
        self.keypointrcnn_loss_customized = keypointrcnn_loss_customized

    @staticmethod
    def set_customized_loss(
        head,
        maskrcnn_loss_customized=None,
        fastrcnn_loss_customized=None,
        keypointrcnn_loss_customized=None,
    ):
        head.maskrcnn_loss_customized = maskrcnn_loss_customized
        head.fastrcnn_loss_customized = fastrcnn_loss_customized
        head.keypointrcnn_loss_customized = keypointrcnn_loss_customized

    @staticmethod
    def update_forward_func(head):
        head.forward = types.MethodType(
            RoIHeads_loss_customized.forward, head
        )  # bound the method to our head

    # def forward(self, features, proposals, image_shapes, targets=None):
    def forward(self, features, proposals, image_shapes, targets=None):
        """
        Arguments:
            features (List[Tensor])
            proposals (List[Tensor[N, 4]])
            image_shapes (List[Tuple[H, W]])
            targets (List[Dict])
        """
        maskrcnn_loss_func = maskrcnn_loss
        fastrcnn_loss_func = fastrcnn_loss
        keypointrcnn_loss_func = keypointrcnn_loss

        eval_when_train = not self.training
        try:
            if self._eval_when_train:
                eval_when_train = True
        except AttributeError:
            pass

        if self.maskrcnn_loss_customized is not None:
            maskrcnn_loss_func = self.maskrcnn_loss_customized
        if self.fastrcnn_loss_customized is not None:
            fastrcnn_loss_func = self.fastrcnn_loss_customized
        if self.keypointrcnn_loss_customized is not None:
            keypointrcnn_loss_func = self.keypointrcnn_loss_customized

        if self.training:
            (
                proposals,
                matched_idxs,
                labels,
                regression_targets,
            ) = self.select_training_samples(proposals, targets)

        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        class_logits, box_regression = self.box_predictor(box_features)

        result, losses = [], {}
        if self.training:
            loss_classifier, loss_box_reg = fastrcnn_loss_func(
                class_logits, box_regression, labels, regression_targets
            )
            losses = dict(loss_classifier=loss_classifier, loss_box_reg=loss_box_reg)
        if eval_when_train:
            boxes, scores, labels = self.postprocess_detections(
                class_logits, box_regression, proposals, image_shapes
            )
            num_images = len(boxes)
            for i in range(num_images):
                result.append(dict(boxes=boxes[i], labels=labels[i], scores=scores[i],))

        if self.has_mask:
            mask_proposals = [p["boxes"] for p in result]
            if self.training:
                # during training, only focus on positive boxes
                num_images = len(proposals)
                mask_proposals = []
                pos_matched_idxs = []
                for img_id in range(num_images):
                    pos = torch.nonzero(labels[img_id] > 0).squeeze(1)
                    mask_proposals.append(proposals[img_id][pos])
                    pos_matched_idxs.append(matched_idxs[img_id][pos])

            mask_features = self.mask_roi_pool(features, mask_proposals, image_shapes)
            mask_features = self.mask_head(mask_features)
            mask_logits = self.mask_predictor(mask_features)

            loss_mask = {}
            if self.training:
                gt_masks = [t["masks"] for t in targets]
                gt_labels = [t["labels"] for t in targets]
                loss_mask = maskrcnn_loss_func(
                    mask_logits, mask_proposals, gt_masks, gt_labels, pos_matched_idxs
                )
                loss_mask = dict(loss_mask=loss_mask)
            if eval_when_train:
                labels = [r["labels"] for r in result]
                masks_probs = maskrcnn_inference(mask_logits, labels)
                for mask_prob, r in zip(masks_probs, result):
                    r["masks"] = mask_prob

            losses.update(loss_mask)

        if self.has_keypoint():
            keypoint_proposals = [p["boxes"] for p in result]
            if self.training:
                # during training, only focus on positive boxes
                num_images = len(proposals)
                keypoint_proposals = []
                pos_matched_idxs = []
                for img_id in range(num_images):
                    pos = torch.nonzero(labels[img_id] > 0).squeeze(1)
                    keypoint_proposals.append(proposals[img_id][pos])
                    pos_matched_idxs.append(matched_idxs[img_id][pos])

            keypoint_features = self.keypoint_roi_pool(
                features, keypoint_proposals, image_shapes
            )
            keypoint_features = self.keypoint_head(keypoint_features)
            keypoint_logits = self.keypoint_predictor(keypoint_features)

            loss_keypoint = {}
            if self.training:
                gt_keypoints = [t["keypoints"] for t in targets]
                loss_keypoint = keypointrcnn_loss_func(
                    keypoint_logits, keypoint_proposals, gt_keypoints, pos_matched_idxs
                )
                loss_keypoint = dict(loss_keypoint=loss_keypoint)
            if eval_when_train:
                keypoints_probs, kp_scores = keypointrcnn_inference(
                    keypoint_logits, keypoint_proposals
                )
                for keypoint_prob, kps, r in zip(keypoints_probs, kp_scores, result):
                    r["keypoints"] = keypoint_prob
                    r["keypoints_scores"] = kps

            losses.update(loss_keypoint)

        return result, losses


def project_masks_on_boxes(gt_masks, boxes, matched_idxs, M):
    """
    Given segmentation masks and the bounding boxes corresponding
    to the location of the masks in the image, this function
    crops and resizes the masks in the position defined by the
    boxes. This prepares the masks for them to be fed to the
    loss computation as the targets.
    """
    matched_idxs = matched_idxs.to(boxes)
    rois = torch.cat([matched_idxs[:, None], boxes], dim=1)
    gt_masks = gt_masks[:, None].to(rois)
    return roi_align(gt_masks, rois, (M, M), 1)[:, 0]


def maskrcnn_loss_focal(
    mask_logits, proposals, gt_masks, gt_labels, mask_matched_idxs, focal_loss_func=None
):
    """
    Arguments:
        proposals (list[BoxList])
        mask_logits (Tensor)
        targets (list[BoxList])

    Return:
        mask_loss (Tensor): scalar tensor containing the loss
    """

    discretization_size = mask_logits.shape[-1]
    labels = [l[idxs] for l, idxs in zip(gt_labels, mask_matched_idxs)]
    mask_targets = [
        project_masks_on_boxes(m, p, i, discretization_size)
        for m, p, i in zip(gt_masks, proposals, mask_matched_idxs)
    ]

    labels = torch.cat(labels, dim=0)
    mask_targets = torch.cat(mask_targets, dim=0)

    # torch.mean (in binary_cross_entropy_with_logits) doesn't
    # accept empty tensors, so handle it separately
    if mask_targets.numel() == 0:
        return mask_logits.sum() * 0

    loss_func = F.binary_cross_entropy_with_logits
    if focal_loss_func is not None:
        loss_func = focal_loss_func

    mask_loss = loss_func(
        mask_logits[torch.arange(labels.shape[0], device=labels.device), labels],
        mask_targets,
    )
    return mask_loss


class TorchModelAnalyzer:
    def __init__(self, kernel):
        self.activation = {}
        self.kernel = kernel

    def get_output_saved(self, name):
        self.activation[name] = []

        def rpn_output_hook(model, input, output):
            if name == "roi_heads":
                self.activation[name].append(self.roi_heads_output_detach(output))

        return rpn_output_hook

    # def put_one_predict(self, name, detached_output):

    def print_output(self, name):
        print(self.activation.get(name, f"output information for {name} not found"))

    def test_out_threshold(self, activations):
        stat_for_threshold = {}
        pred_for_threshold = {}
        for threshold in [0.5, 0.52, 0.55, 0.6]:
            preds, stat = self.kernel.predict_rle_from_acts_with_threshold(
                activations, threshold
            )

            stat_for_threshold[threshold] = stat
            pred_for_threshold[threshold] = preds
            # my_trace()
            # print(self.metric_cal(preds))

        kernel_utils.dump_obj(stat_for_threshold, "stat_for_threshold.pkl", force=True)
        kernel_utils.dump_obj(pred_for_threshold, "pred_for_threshold.pkl", force=True)

    def metric_cal(self, preds):
        pass

    @staticmethod
    def roi_heads_output_detach(output):
        return (
            [
                {k: v.detach().cpu().numpy() for k, v in dict.items()}
                for dict in output[0]
            ],
            {k: v.detach().cpu().numpy() for k, v in output[1].items()},
        )

    @staticmethod
    def register_forward_hook(module, func_to_hook):
        module.register_forward_hook(func_to_hook)



import logging
from abc import ABCMeta, abstractmethod

from kaggle_runner.kernels.KernelRunningState import KernelRunningState
from kaggle_runner.utils import kernel_utils

# Plot inline
# %matplotlib inline


class KernelGroup:
    """Kernel Group to try different combination of kernels hyperparameter

    Follow this example:
    https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html

    Attributes:
        kernels (list): List of kernels whose results will be analyse together.

    """

    def __init__(self, kernels):
        self.kernels = kernels

class KaggleKernel(metaclass=ABCMeta):
    """Kernel for kaggle competitions/researchs

    Attributes:
        _stage
        model_metrics
        data_loader
        optimizer
        model
        train_X
        developing
        submit_run
        test_X
        model_loss
        num_epochs
        device
        result_analyzer
        logger
        dependency
        dev_X
        train_Y
        dev_Y

    """
    def __init__(self, logger=None):
        self.submit_run = False
        self.num_epochs = 8
        self.device = None
        self.optimizer = None
        self.data_loader = None
        self.developing = True
        self.model = None
        self.model_metrics = []
        self.model_loss = None

        self.train_X = None
        self.train_Y = None

        self.dev_X = None
        self.dev_Y = None

        self.test_X = None

        self.result_analyzer = None  # for analyze the result

        self._stage = KernelRunningState.INIT_DONE
        self.logger = logger
        self.dependency = []

    def _add_dependency(self, dep):
        """_add_dependency just install pip dependency now
        """
        self.dependency.append(dep)

    def install_dependency(self, dep):
        self._add_dependency(dep)

    def _add_logger_handler(self, handler):
        self.logger.addHandler(handler)

    def set_logger(self, name, level=logging.DEBUG, handler=None):
        FORMAT = "[%(levelname)s]%(asctime)s:%(name)s:%(message)s"
        logging.basicConfig(format=FORMAT)
        logger = logging.getLogger(name)
        logger.setLevel(level)

        if handler is not None:
            logger.addHandler(handler)
        self.logger = logger

    def set_random_seed(self):
        pass

    def set_data_size(self):
        "might be useful when test different input datasize"
        pass

    def save_model(self):
        pass

    def load_model_weight(self):
        pass

    @abstractmethod
    def build_and_set_model(self):
        pass

    def train_model(self):
        pass

    def set_model(self):
        pass

    def set_loss(self, loss_func):
        pass

    def set_metrics(self, metrics):
        """
        set_metrics for model training

        :return: None
        """

    def set_result_analyzer(self):
        pass

    def pre_prepare_data_hook(self):
        pass

    def after_prepare_data_hook(self):
        pass

    def prepare_train_dev_data(self):
        pass

    def prepare_test_data(self, data_config=None):
        pass

    @abstractmethod
    def peek_data(self):
        pass

    def predict_on_test(self):
        pass

    def dump_state(self, exec_flag=False):
        self.logger.debug("state %s" % self._stage)

        if exec_flag:
            self.logger.debug("dumping state to file for %s" % self._stage)
            # dump_obj(self, 'run_state.pkl', force=True)  # too large
            kernel_utils.dump_obj(self, "run_state_%s.pkl" %
                                  self._stage, force=True)

    def run(
        self,
        start_stage=None,
        end_stage=KernelRunningState.SAVE_SUBMISSION_DONE,
        dump_flag=False,
    ):
        """

        :param start_stage: if set, will overwrite the stage
        :param end_stage:
        :param dump_flag:
        :return:
        """
        self.continue_run(
            start_stage=start_stage, end_stage=end_stage, dump_flag=dump_flag
        )

    def continue_run(
        self,
        start_stage=None,
        end_stage=KernelRunningState.SAVE_SUBMISSION_DONE,
        dump_flag=False,
    ):
        self.set_random_seed()
        self.logger.debug(
            "%s -> %s", start_stage, end_stage,
        )

        if start_stage is not None:
            assert start_stage.value < end_stage.value
            self._stage = start_stage

        if self._stage.value < KernelRunningState.PREPARE_DATA_DONE.value:
            self.pre_prepare_data_hook()
            self.prepare_train_dev_data()
            self.after_prepare_data_hook()

            self._stage = KernelRunningState.PREPARE_DATA_DONE
            self.dump_state(exec_flag=dump_flag)

            if self._stage.value >= end_stage.value:
                return

        if self._stage.value < KernelRunningState.TRAINING_DONE.value:
            self.pre_train()
            self.build_and_set_model()
            self.train_model()
            self.after_train()

            self.save_model()  # during training, it will also save model

            self._stage = KernelRunningState.TRAINING_DONE
            self.dump_state(exec_flag=dump_flag)

            if self._stage.value >= end_stage.value:
                return

        if self._stage.value < KernelRunningState.EVL_DEV_DONE.value:
            self.set_result_analyzer()

            self._stage = KernelRunningState.EVL_DEV_DONE
            self.dump_state(exec_flag=dump_flag)

            if self._stage.value >= end_stage.value:
                return

        if self._stage.value < KernelRunningState.SAVE_SUBMISSION_DONE.value:
            self.pre_test()
            self.prepare_test_data()
            self.predict_on_test()
            self.after_test()

            self.pre_submit()
            self.submit()
            self.after_submit()

            self._stage = KernelRunningState.SAVE_SUBMISSION_DONE
            self.dump_state(exec_flag=dump_flag)

            if self._stage.value >= end_stage.value:
                return

    @classmethod
    def _load_state(cls, stage=None, file_name="run_state.pkl", logger=None):
        """

        :param file_name:
        :return: the kernel object, need to continue
        """

        if stage is not None:
            file_name = f"run_state_{stage}.pkl"

        if logger is not None:
            logger.debug(f"restore from {file_name}")
        self = kernel_utils.get_obj_or_dump(filename=file_name)
        assert self is not None
        self.logger = logger

        return self

    def load_state_data_only(self, file_name="run_state.pkl"):
        pass

    @classmethod
    def load_state_continue_run(cls, file_name="run_state.pkl", logger=None):
        """

        :param file_name:
        :return: the kernel object, need to continue
        """
        self = cls._load_state(file_name=file_name, logger=logger)
        self.continue_run()

    def pre_train(self):
        pass

    def after_train(self):
        pass

    def pre_submit(self):
        pass

    def submit(self):
        pass

    def after_submit(self):
        "after_submit should report to our logger, for next step analyze"
        pass

    def pre_test(self):
        pass

    def after_test(self):
        pass

    def save_result(self):
        pass

    def plot_train_result(self):
        pass

    def plot_test_result(self):
        pass

    def analyze_data(self):
        pass

    @abstractmethod
    def check_predict_details(self):
        pass

from kaggle_runner import logger

class KaggleKernelOnlyPredict(KaggleKernel):

    def __init__(self, model_path):
        super(KaggleKernelOnlyPredict, self).__init__(logger=logger)
        self.only_predict = True

    @abstractmethod
    def build_and_set_model(self):
        """load pretrained one"""
        pass

    def prepare_train_dev_data(self):
        pass

    @abstractmethod
    def prepare_test_data(self, data_config=None):
        pass

    @abstractmethod
    def check_predict_details(self):
        pass

    def peek_data(self):
        pass


def test_init_only_predict():
    k = KaggleKernelOnlyPredict()
    assert k is not None
    assert k.model is not None



from unittest import TestCase


class TestKaggleKernel(TestCase):
    def test_install_dependency(self):
        self.fail()

    def test_set_logger(self):
        self.fail()

    def test_set_random_seed(self):
        self.fail()

    def test_save_model(self):
        self.fail()

    def test_train_model(self):
        self.fail()

    def test_set_model(self):
        self.fail()

    def test_prepare_train_dev_data(self):
        self.fail()

    def test_prepare_test_data(self):
        self.fail()

    def test_run(self):
        self.fail()

    def test_continue_run(self):
        self.fail()

    def test_submit(self):
        self.fail()



import os
from kaggle_runner import logger
from kaggle_runner.kernels.fastai_kernel import FastAIKernel
from kaggle_runner.metrics.metrics import matthews_correlation
from kaggle_runner.datasets.transfomers import *
from kaggle_runner.datasets.bert import DatasetRetriever
from kaggle_runner.utils.kernel_utils import get_obj_or_dump
from kaggle_runner.modules.ToxicSimpleNNModel import ToxicSimpleNNModel
from fastai.basic_data import DataBunch
import transformers
from transformers import *
import albumentations

ROOT_PATH = f'/kaggle' # for colab
BACKBONE_PATH = 'xlm-roberta-large'

from kaggle_runner.defaults import SEED
import random
import numpy as np

def seed_everything(seed):
    """seed_everything.

    :param seed: number to generate pseudo random states
    """
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True

def get_train_transforms():
    """get_train_transforms."""

    return albumentations.Compose([
        ExcludeUsersMentionedTransform(p=0.95),
        ExcludeUrlsTransform(p=0.95),
        ExcludeNumbersTransform(p=0.95),
        ExcludeHashtagsTransform(p=0.95),
        ExcludeDuplicateSentencesTransform(p=0.95),
    ], p=1.0)

def get_synthesic_transforms(supliment_toxic, p=0.5, mix=False):
    """get_synthesic_transforms.

    :param supliment_toxic: extra toxic data which will be used in synthesizing data
    :param p: probability
    :param mix: flag for mixing more data with toxic (it does not work with balanced sampler, too many toxic data)
    """

    return SynthesicOpenSubtitlesTransform(p=p, supliment_toxic=supliment_toxic, mix=mix)

ROOT_PATH = '/kaggle'
def get_pickled_data(file_path):
    """get_pickled_data from current folder or kaggle data input folder.

    :param file_path:
    """
    obj = get_obj_or_dump(file_path)

    if obj is None:
        #may_debug(True)

        return get_obj_or_dump(f"{ROOT_PATH}/input/clean-pickle-for-jigsaw-toxicity/{file_path}")

    return obj

vocab = get_pickled_data("vocab.pkl")
#if vocab is None: # vocab file read~~
#   vocab = [tokenizer.convert_ids_to_tokens(i) for i in range(tokenizer.vocab_size)]
#   get_obj_or_dump("vocab.pkl", default=vocab)

class Shonenkov(FastAIKernel):
    """Shonenkov kernel, use TPU to train; for data, down sampler and synthesizing."""

    def __init__(self, device, config, **kargs):
        """__init__.

        :param device:
        :param config:
        :param kargs:
        """
        super(Shonenkov, self).__init__(**kargs)
        self.data = None
        self.transformers = None
        self.setup_transformers()
        self.device = device
        self.config = config
        self.learner = None

    def build_and_set_model(self):
        """build_and_set_model."""
        self.model = ToxicSimpleNNModel()
        self.model = self.model.to(self.device)

    def set_random_seed(self):
        """set_random_seed."""
        seed_everything(SEED)

    def setup_transformers(self):
        """setup_transformers."""

        if self.transformers is None:
            supliment_toxic = None # avoid overfit
            train_transforms = get_train_transforms();
            synthesic_transforms_often = get_synthesic_transforms(supliment_toxic, p=0.5)
            synthesic_transforms_low = None
            tokenizer = XLMRobertaTokenizer.from_pretrained(BACKBONE_PATH)
            shuffle_transforms = ShuffleSentencesTransform(always_apply=True)

            self.transformers = {'train_transforms': train_transforms,
                                 'synthesic_transforms_often': synthesic_transforms_often,
                                 'synthesic_transforms_low': synthesic_transforms_low,
                                 'tokenizer': tokenizer, 'shuffle_transforms':
                                 shuffle_transforms}

    def prepare_train_dev_data(self):
        """prepare_train_dev_data."""
        df_train = get_pickled_data("train.pkl")

        if df_train is None:
            df_train = pd.read_csv(f'{ROOT_PATH}/input/jigsaw-toxicity-train-data-with-aux/train_data.csv')
            df_train['comment_text'] = df_train.parallel_apply(lambda x: clean_text(x['comment_text'], x['lang']), axis=1)
            get_obj_or_dump("train.pkl", default=df_train)

        #supliment_toxic = get_toxic_comments(df_train)
        self.train_dataset = DatasetRetriever(
            labels_or_ids=df_train['toxic'].values,
            comment_texts=df_train['comment_text'].values,
            langs=df_train['lang'].values,
            severe_toxic=df_train['severe_toxic'].values,
            obscene=df_train['obscene'].values,
            threat=df_train['threat'].values,
            insult=df_train['insult'].values,
            identity_hate=df_train['identity_hate'].values,
            use_train_transforms=True,
            transformers=self.transformers
        )
        df_val = get_pickled_data("val.pkl")

        if df_val is None:
            df_val = pd.read_csv(f'{ROOT_PATH}/input/jigsaw-multilingual-toxic-comment-classification/validation.csv', index_col='id')
            df_val['comment_text'] = df_val.parallel_apply(lambda x: clean_text(x['comment_text'], x['lang']), axis=1)
            get_obj_or_dump("val.pkl", default=df_val)

        self.validation_tune_dataset = DatasetRetriever(
            labels_or_ids=df_val['toxic'].values,
            comment_texts=df_val['comment_text'].values,
            langs=df_val['lang'].values,
            use_train_transforms=True,
            transformers=self.transformers
        )
        self.validation_dataset = DatasetRetriever(
            labels_or_ids=df_val['toxic'].values,
            comment_texts=df_val['comment_text'].values,
            langs=df_val['lang'].values,
            use_train_transforms=False,
            transformers=self.transformers
        )

        del df_val
        gc.collect();

        del df_train
        gc.collect();

    def prepare_test_data(self):
        """prepare_test_data."""

        if os.path.exists('/content'): # colab
            df_test = get_pickled_data("test.pkl")
        else:
            df_test = None

        if df_test is None:
            df_test = pd.read_csv(f'{ROOT_PATH}/input/jigsaw-multilingual-toxic-comment-classification/test.csv', index_col='id')
            df_test['comment_text'] = df_test.parallel_apply(lambda x: clean_text(x['content'], x['lang']), axis=1)
            get_obj_or_dump("test.pkl", default=df_test)

        self.test_dataset = DatasetRetriever(
            labels_or_ids=df_test.index.values, ## here different!!!
            comment_texts=df_test['comment_text'].values,
            langs=df_test['lang'].values,
            use_train_transforms=False,
            test=True,
            transformers=self.transformers
        )

        del df_test
        gc.collect();
    def after_prepare_data_hook(self):
        """Put to databunch here"""
        logger.debug("kernel use device %s", self.device)
        self.data = DataBunch.create(self.train_dataset,
                                     self.validation_dataset,
                                     bs=self.config.batch_size,
                                     device=self.device,
                                     num_workers=self.config.num_workers)

    def peek_data(self):
        """peek_data."""

        if self.data is not None:
            may_debug()
            o = self.data.one_batch()
            print(o)

            return o
        else:
            if self.logger is not None:
                self.logger.error("peek_data failed, DataBunch is None.")

from kaggle_runner import may_debug

import torch.nn as nn

class ToxicSimpleNNModelChangeInner(nn.Module):
    """ToxicSimpleNNModelChangeInner."""


    def __init__(self, use_aux=True):
        """__init__.

        :param use_aux:
        """
        super(ToxicSimpleNNModelChangeInner, self).__init__()

        self.backbone = XLMModel.from_pretrained('xlm-mlm-tlm-xnli15-1024')
        self.dropout = nn.Dropout(0.3)
        aux_len = 0

        if use_aux:
            aux_len = 5

        #in_features = self.backbone.layer[11].ff.layer_2.out_features*2
        #in_features = self.backbone.pooler.dense.out_features*2 # bert - xmlrobert
        in_features = self.backbone.ffns[11].lin2.out_features*2 # xlm
        self.linear = nn.Linear(
            in_features=in_features,
            out_features=2+aux_len,
        )

    def forward(self, input_ids, attention_masks):
        """forward.

        :param input_ids:
        :param attention_masks:
        """
        bs, seq_length = input_ids.shape
        seq_x, _ = self.backbone(
            input_ids=input_ids, attention_mask=attention_masks)
        apool = torch.mean(seq_x, 1)
        mpool, _ = torch.max(seq_x, 1)
        x = torch.cat((apool, mpool), 1)
        x = self.dropout(x)

        return self.linear(x)

class ShonenkovChangeInner(Shonenkov):
    """ShonenkovChangeInner."""

    def __init__(self, device, config, **kargs):
        """__init__.

        :param device:
        :param config:
        :param kargs:
        """
        super(ShonenkovChangeInner, self).__init__(device, config, **kargs)
        assert self.transformers is not None

    def build_and_set_model(self):
        """build_and_set_model."""
        self.model = ToxicSimpleNNModelChangeInner()
        self.model = self.model.to(self.device)

    def prepare_train_dev_data(self):
        """prepare_train_dev_data."""
        df_train = get_pickled_data("train_XLM.pkl")

        if df_train is None:
            df_train = pd.read_csv(f'{ROOT_PATH}/input/jigsaw-toxicity-train-data-with-aux/train_data.csv')
            df_train['comment_text'] = df_train.parallel_apply(lambda x: clean_text(x['comment_text'], x['lang']), axis=1)
            get_obj_or_dump("train_XLM.pkl", default=df_train)

        #supliment_toxic = get_toxic_comments(df_train)
        self.train_dataset = DatasetRetriever(
            labels_or_ids=df_train['toxic'].values,
            comment_texts=df_train['comment_text'].values,
            langs=df_train['lang'].values,
            severe_toxic=df_train['severe_toxic'].values,
            obscene=df_train['obscene'].values,
            threat=df_train['threat'].values,
            insult=df_train['insult'].values,
            identity_hate=df_train['identity_hate'].values,
            use_train_transforms=True,
            transformers=self.transformers
        )
        df_val = get_pickled_data("val_XLM.pkl")

        if df_val is None:
            df_val = pd.read_csv(f'{ROOT_PATH}/input/jigsaw-multilingual-toxic-comment-classification/validation.csv', index_col='id')
            df_val['comment_text'] = df_val.parallel_apply(lambda x: clean_text(x['comment_text'], x['lang']), axis=1)
            get_obj_or_dump("val_XLM.pkl", default=df_val)

        self.validation_tune_dataset = DatasetRetriever(
            labels_or_ids=df_val['toxic'].values,
            comment_texts=df_val['comment_text'].values,
            langs=df_val['lang'].values,
            use_train_transforms=True,
            transformers=self.transformers
        )
        self.validation_dataset = DatasetRetriever(
            labels_or_ids=df_val['toxic'].values,
            comment_texts=df_val['comment_text'].values,
            langs=df_val['lang'].values,
            use_train_transforms=False,
            transformers=self.transformers
        )

        del df_val
        gc.collect();

        del df_train
        gc.collect();

    def prepare_test_data(self):
        """prepare_test_data."""

        if os.path.exists('/content'): # colab
            df_test = get_pickled_data("test_XLM.pkl")
        else:
            df_test = None

        if df_test is None:
            df_test = pd.read_csv(f'{ROOT_PATH}/input/jigsaw-multilingual-toxic-comment-classification/test.csv', index_col='id')
            df_test['comment_text'] = df_test.parallel_apply(lambda x: clean_text(x['content'], x['lang']), axis=1)
            get_obj_or_dump("test_XLM.pkl", default=df_test)

        self.test_dataset = DatasetRetriever(
            labels_or_ids=df_test.index.values, ## here different!!!
            comment_texts=df_test['comment_text'].values,
            langs=df_test['lang'].values,
            use_train_transforms=False,
            test=True,
            transformers=self.transformers
        )

        del df_test
        gc.collect();

    def setup_transformers(self):
        """setup_transformers."""

        if not hasattr(self, 'transformers') or self.transformers is None:
            supliment_toxic = None # avoid overfit
            train_transforms = get_train_transforms();
            synthesic_transforms_often = get_synthesic_transforms(supliment_toxic, p=0.5)
            synthesic_transforms_low = None
            shuffle_transforms = ShuffleSentencesTransform(always_apply=True)

            #from tokenizers import BertWordPieceTokenizer
            #tokenizer = XLNetTokenizer.from_pretrained('xlnet-base-cased')
            tokenizer = XLMTokenizer.from_pretrained('xlm-mlm-tlm-xnli15-1024')

            self.transformers = {'train_transforms': train_transforms,
                                 'synthesic_transforms_often': synthesic_transforms_often,
                                 'synthesic_transforms_low': synthesic_transforms_low,
                                 'tokenizer': tokenizer,
                                 'shuffle_transforms': shuffle_transforms}

import torch
class DummyTrainGlobalConfig:
    """ Global Config for this notebook """
    num_workers = 0  # количество воркеров для loaders
    batch_size = 16  # bs , 8 for GPU, 16 for TPU
    n_epochs = 2  # количество эпох для обучения
    lr = 0.3 * 1e-5 # стартовый learning rate (внутри логика работы с мульти TPU домножает на кол-во процессов)
    fold_number = 0  # номер фолда для обучения

    # -------------------
    verbose = True  # выводить принты
    verbose_step = 25  # количество шагов для вывода принта
    # -------------------

    # --------------------
    step_scheduler = False  # выполнять scheduler.step после вызова optimizer.step
    validation_scheduler = True  # выполнять scheduler.step после валидации loss (например для плато)
    SchedulerClass = torch.optim.lr_scheduler.ReduceLROnPlateau
    scheduler_params = dict(
        mode='max',
        factor=0.7,
        patience=0,
        verbose=False,
        threshold=0.0001,
        threshold_mode='abs',
        cooldown=0,
        min_lr=1e-8,
        eps=1e-08
    )
    # --------------------

    # -------------------
    criterion = 'L'
    # -------------------


def test_change_XLM_module():
    """test_change_XLM_module."""
    from kaggle_runner.losses import LabelSmoothing
    k = ShonenkovChangeInner(torch.device("cpu"), DummyTrainGlobalConfig,
                             metrics=None, loss_func=LabelSmoothing(),
                             opt_func=None)
    assert k is not None
    k.run(dump_flag=False)



#!/bin/env python
import csv
import os
import random
import re
import subprocess

import pandas as pd

from kaggle_runner import logger

DATA_PATH = "/kaggle/input/jigsaw-multilingual-toxic-comment-classification"
TRD = "jigsaw-toxic-comment-train.csv"
VD = "validation.csv"
TD = "test.csv"
OUT_PATH = "/tmp/input.txt"


def get_column(csv_file, col_name):
    d = pd.read_csv(csv_file)

    return d[col_name]

def get_toxic_comment(p, col_name="comment_text"):
    dtr = get_column(os.path.join(DATA_PATH, p), col_name=col_name)

    r = random.randint(0, len(dtr)-1)
    logger.debug("toxic data example %s.[%s]:\n %s\n", p, col_name, dtr[r])

    return dtr

def join_line(ml):
    return ml.replace("\n", "\\n")

def clean(text):
    text = text.fillna("fillna")
    text = text.map(lambda x: re.sub('\\n', ' ', str(x)))
    text = text.map(lambda x: re.sub(
        "\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", '', str(x)))
    text = text.map(lambda x: re.sub(
        "\(http://.*?\s\(http://.*\)", '', str(x)))

    return text

def merge_all_data():
    dtr = get_toxic_comment(TRD)
    vd = get_toxic_comment(VD)
    td = get_toxic_comment(TD, "content")
    dtr =clean(dtr)
    vd = clean( vd)
    td = clean( td)
    c = pd.concat([dtr, vd, td])
    pd.DataFrame.to_csv(c,OUT_PATH, index=False, header=False)
    print("%d %d %d, %d" % (len(dtr), len(vd), len(td), sum((len(dtr), len(vd), len(td))))) # lines info: 223549 8000 63812


if __name__ == "__main__":
    merge_all_data()



import warnings
warnings.simplefilter('ignore')

from fastai import *
from fastai.vision import *
from fastai.vision import get_transforms

import os

from pathlib import Path

import pandas as pd

INPUT = Path("../input/digit-recognizer")
os.listdir(INPUT)

train_df = pd.read_csv(INPUT/"train.csv")

test_df = pd.read_csv(INPUT/"test.csv")

TRAIN = Path("../train")
TEST = Path("../test")

for index in range(10):
    try:
        os.makedirs(TRAIN/str(index))
    except:
        pass

sorted(os.listdir(TRAIN))

try:
    os.makedirs(TEST)
except:
    pass

from PIL import Image

def saveDigit(digit, filepath):
    digit = digit.reshape(28,28)
    digit = digit.astype(np.uint8)

    img = Image.fromarray(digit)
    img.save(filepath)


for index, row in train_df.iterrows():

    label,digit = row[0], row[1:]

    folder = TRAIN/str(label)
    filename = f"{index}.jpg"
    filepath = folder/filename

    digit = digit.values

    saveDigit(digit, filepath)


for index, digit in test_df.iterrows():

    folder = TEST
    filename = f"{index}.jpg"
    filepath = folder/filename

    digit = digit.values

    saveDigit(digit, filepath)


import matplotlib.pyplot as plt

def displayTrainingData():
    fig = plt.figure(figsize=(5,10))

    for rowIndex in range(1, 10):
        subdirectory = str(rowIndex)
        path = TRAIN/subdirectory
        images = os.listdir(path)

        for sampleIndex in range(1, 6):
            randomNumber = random.randint(0, len(images)-1)
            image = Image.open(path/images[randomNumber])
            ax = fig.add_subplot(10, 5, 5*rowIndex + sampleIndex)
            ax.axis("off")

            plt.imshow(image, cmap='gray')

    plt.show()

def displayTestingData():
    fig = plt.figure(figsize=(5, 10))

    paths = os.listdir(TEST)


    for i in range(1, 51):
        randomNumber = random.randint(0, len(paths)-1)
        image = Image.open(TEST/paths[randomNumber])

        ax = fig.add_subplot(10, 5, i)
        ax.axis("off")

        plt.imshow(image, cmap='gray')
    plt.show()

print('samples of training data')
displayTrainingData()

print('samples of testing data')
displayTestingData()


image_path = TEST/os.listdir(TEST)[9]
image = Image.open(image_path)
image_array = np.asarray(image)


#fig, ax = plt.subplots(figsize=(15, 15))

#img = ax.imshow(image_array, cmap='gray')

#for x in range(28):
#    for y in range(28):
#        value = round(image_array[y][x]/255.0, 2)
#        color = 'black' if value > 0.5 else 'white'
#        ax.annotate(s=value, xy=(x, y), ha='center', va='center', color=color)
#
#plt.axis('off')
#plt.show()


tfms = get_transforms(do_flip=False)

data = ImageDataBunch.from_folder(
    path = str(TRAIN),
    test = str(TEST),
    valid_pct = 0.2,
    bs = 16,
    size = 28,
    #num_workers = 0,
    ds_tfms = tfms
)


print(mnist_stats)

data.normalize(mnist_stats)

print(data.classes)





learn = cnn_learner(data, base_arch=models.resnet18, metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)





learn.fit_one_cycle(cyc_len=5)


interp = ClassificationInterpretation.from_learner(learn)


interp.plot_top_losses(9, figsize=(7, 7))


interp.plot_confusion_matrix()













import platform; platform.system()


flip_tfm = RandTransform(tfm=TfmPixel (flip_lr), kwargs={}, p=1, resolved={}, do_run=True, is_random=True, use_on_y=True)
folder = TRAIN/"3"
filename = os.listdir(folder)[0]
img = open_image(TRAIN/folder/filename)
display(img)
display(img.apply_tfms(flip_tfm))


tfms = get_transforms(do_flip=False)







learn = cnn_learner(data, base_arch=models.densenet169, metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)


learn = cnn_learner(data, base_arch=models.densenet169, pretrained=False, metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)



import torchvision.models


learn = Learner(data, torchvision.models.googlenet(), metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)


learn = Learner(data, torchvision.models.googlenet(pretrained=True), metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)




class CNN(nn.Module):
    def __init__(self):
        super(CNN, self).__init__()
        # here you instantiate all the layers of the neural network and the activation function

    def forward(self, x):
        # here you define the forward propagation

        return x



batch_size = 16

class CNN(nn.Module):
    def __init__(self):
        super(CNN, self).__init__()
        # input is 28 pixels x 28 pixels x 3 channels
        # our original data was grayscale, so only one channel, but fast.ai automatically loads in the data as RGB
        self.conv1 = nn.Conv2d(3,16, 3, padding=1)
        self.conv2 = nn.Conv2d(16,32, 3, padding=1)
        self.pool = nn.MaxPool2d(2, 2)
        self.fc1 = nn.Linear(7*7*32, 500)
        self.fc2 = nn.Linear(500, 10)
        self.relu = nn.ReLU()

    def forward(self, x):
        print (x.size())
        # x (28x28x3)
        x = self.conv1(x)
        # x (28x28x16)
        x = self.pool(x)
        # x (14x14x16)
        x = self.relu(x)

        x = self.conv2(x)
        # x (14x14x32)
        x = self.pool(x)
        # x (7x7x32)
        x = self.relu(x)

        # flatten images in batch
        print(x.size())
        x = x.view(-1,7*7*32)
        print(x.size())
        x = self.fc1(x)
        x = self.relu(x)

        x = self.fc2(x)
        x = self.relu(x)

        return x

learn = Learner(data, CNN(), metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)



import subprocess
import os
import sys
__all__ = ["may_debug", "logger", "runners.runner",
           "runners.coordinator", "utils"]

from .defaults import DEBUG, RIPDB, INTERACTIVE

from .utils import logger

logger = logger

def reload_me():
    from importlib import reload
    current_module = sys.modules[__name__]
    reload(current_module)

def may_debug(force=False):
    subprocess.run('cd /kaggle/working || cd kaggle_runner; '
        'git commit -asm "Good game"; git pull; git submodule update --init --recursive', shell=True)
    reload_me()

    import pdb

    if force:
        if "pytest" in sys.modules:
            import pytest
            pytest.set_trace()
        else:
            if INTERACTIVE:
                import ipdb
                ipdb.set_trace()
            else:
                pdb.set_trace()
    elif DEBUG:
        if "pytest" in sys.modules:
            import pytest
            pytest.set_trace()
        else:
            if INTERACTIVE:
                import ipdb
                ipdb.set_trace()
            else:
                pdb.set_trace()



import os
import shutil
import subprocess
import sys

import pytest

from kaggle_runner import utils
from kaggle_runner.runners import coordinator


@pytest.fixture(scope="module")
def runner_configs():
    return [
        {"port":23454, "size": 384, "network": "intercept", "AMQPURL": utils.AMQPURL()},
        {"port":23454, "size": 384, "network": "intercept-resnet", "AMQPURL": utils.AMQPURL()},
    ]


class TestCoordinator:
    coordinator = None
    tmp_path = "."

    @classmethod
    def setup_class(cls):
        cls.tmp_path = ".runners"
        cls.coordinator = coordinator.Coordinator(cls.tmp_path, "Test Runner")
        print("setup_class called once for the class")

    @classmethod
    def teardown_class(cls):
        print("teardown_class called once for the class")

    def setup_method(self, method):
        if os.path.exists(self.tmp_path):
            shutil.rmtree(self.tmp_path)
            os.mkdir(self.tmp_path)
        print("setup_method called for every method")

    def teardown_method(self, method):
        # shutil.rmtree(self.tmp_path)  # for debug
        print("teardown_method called for every method")

    def test_generate_runner(self, runner_configs):
        self.coordinator.create_runner(runner_configs[1], 19999, False)
        # ret = self.coordinator.run_local(path)
        # assert ret.returncode == 0

    @pytest.mark.timeout(15)
    def test_push_runner_nb(self, runner_configs):
        path = self.coordinator.create_runner(runner_configs[1], 19999, False)
        # ret = self.coordinator.run_local(path)
        # assert ret.returncode == 0

        if os.getenv("CI") != "true":
            ret = self.coordinator.push(path)  # just push first
            assert ret.returncode == 0

    def test_push_runner_cmd(self, runner_configs):
        subprocess.run(f"python -m kaggle_runner "
                       f"{runner_configs[1]['port']} dev", shell=True, check=True)

    @pytest.mark.timeout(10)
    @pytest.mark.skip("runner runs in computation server, no need test local")
    def test_get_mq_back(self, runner_configs):
        path = self.coordinator.create_runner(runner_configs[1], 20202)
        ret = self.coordinator.push(path)
        assert ret.returncode == 0
        # just use a timeout, not within then return error
        self.coordinator._get_result(timeout=100)

    @pytest.mark.skip("runner runs in computation server, no need test local")
    def test_create_runners(self, runner_configs):
        """Should just use unit test setup and teardown
        """

        for c in runner_configs:
            r = self.coordinator.create_runner(c)  # we need to let it run
        assert r.AMQPURL is not None


class TestMain:
    def test_call_remote_mq(self):
        call_params = [
            "python",
            "main.py",
            "amqp://drdsfaew:QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI@termite.rmq.cloudamqp.com/drdsfaew",
            "384",  # size 256+128
            "123",
            "intercept-resnet",
        ]
        utils.logger.debug(" ".join(call_params))
        ret = subprocess.run(call_params)
        assert ret.returncode == 0

    @pytest.mark.skip("test done")
    def test_call_local(self):
        call_params = [
            "python",
            "main.py",
            "amqp://guest:guest@127.0.0.1/",
            "384",  # size 256+128
            "123",
            "intercept-resnet",
        ]
        utils.logger.debug(" ".join(call_params))
        ret = subprocess.run(call_params)
        assert ret.returncode == 0



import tensorflow as tf

from lstm import LSTM_UNITS, DENSE_HIDDEN_UNITS


def reinitLayers(model):
    # reinitLayers refer to https://github.com/keras-team/keras/issues/341
    session = K.get_session()
    for layer in model.layers:
        # if isinstance(layer, keras.engine.topology.Container):
        if isinstance(layer, tf.keras.Model):
            reinitLayers(layer)
            continue
        print("LAYER::", layer.name)
        if layer.trainable == False:
            continue
        for v in layer.__dict__:
            v_arg = getattr(layer, v)
            if hasattr(
                v_arg, "initializer"
            ):  # not work for layer wrapper, like Bidirectional
                initializer_method = getattr(v_arg, "initializer")
                initializer_method.run(session=session)
                print("reinitializing layer {}.{}".format(layer.name, v))


def identity_model(features, labels, mode, params):
    # words = Input(shape=(None,))
    embedding_matrix = params["embedding_matrix"]
    identity_out_num = params["identity_out_num"]

    words = features  # input layer, so features need to be tensor!!!

    x = Embedding(*embedding_matrix.shape, weights=[embedding_matrix], trainable=False)(
        words
    )
    x = SpatialDropout1D(0.2)(x)
    x = Bidirectional(CuDNNLSTM(LSTM_UNITS, return_sequences=True))(x)

    hidden = concatenate([GlobalMaxPooling1D()(x), GlobalAveragePooling1D()(x),])
    hidden = add([hidden, Dense(DENSE_HIDDEN_UNITS, activation="relu")(hidden)])

    # Output Layer
    result = Dense(identity_out_num, activation="sigmoid")(hidden)

    # Implement training, evaluation, and prediction
    # Compute predictions.
    # predicted_classes = tf.argmax(input=logits, axis=1)

    if mode == tf.estimator.ModeKeys.PREDICT:
        # predictions = {
        #    'class_ids': predicted_classes[:, tf.newaxis],
        #    'probabilities': tf.nn.softmax(logits),
        #    'logits': logits,
        # }
        return tf.estimator.EstimatorSpec(mode, predictions=result.out)

    # Compute loss.
    loss = tf.keras.losses.binary_crossentropy(
        labels, result
    )  # todo put it together, fine?

    # Compute evaluation metrics.
    # m = tf.keras.metrics.SparseCategoricalAccuracy()
    # m.update_state(labels, logits)
    # accuracy = m.result()
    # metrics = {'accuracy': accuracy}
    # tf.compat.v1.summary.scalar('accuracy', accuracy)
    binary_accuracy = tf.keras.metrics.binary_accuracy(labels, result)

    metrics = {"accuracy": binary_accuracy}
    tf.compat.v1.summary.scalar("accuracy", binary_accuracy[1])

    if mode == tf.estimator.ModeKeys.EVAL:
        return tf.estimator.EstimatorSpec(mode, loss=loss, eval_metric_ops=metrics)

    # Create training op. # will be called by the estimator
    assert mode == tf.estimator.ModeKeys.TRAIN

    # optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=params['learning_rate'])
    optimizer = tf.compat.v1.train.AdamOptimizer(
        learning_rate=tf.compat.v1.train.exponential_decay(
            learning_rate=params["learning_rate"],
            global_step=tf.compat.v1.train.get_global_step(),
            decay_steps=params["decay_steps"],
            staircase=True,
            decay_rate=0.5,
        )
    )
    train_op = optimizer.minimize(
        loss, global_step=tf.compat.v1.train.get_global_step()
    )
    return tf.estimator.EstimatorSpec(mode, loss=loss, train_op=train_op)


from kaggle_runner.utils import utils
from kaggle_runner.utils.utils import parse_AMQP


class Runner:  # blade runner
    """Runner should run in docker container, and then a controller(or just mq) will check the running status
    """
    def __init__(self, kernel_code="TestKernel", AMQPURL=None, **kwargs):
        self.kernel_name = kernel_code
        self.AMQPURL = parse_AMQP(AMQPURL)
        self.logger = None
        self.trainer = None

    def _attach_data_collector(self, kernel=""):
        """
        Credits: https: // github.com/albertomr86/python-logging-rabbitmq

        !pip install python_logging_rabbitmq

        """
        logger = utils.get_logger(self.kernel_name + "_runner")
        self.logger = utils.attach_data_collector(logger, self.AMQPURL)



import pandas as pd
from sklearn.model_selection import StratifiedKFold
from torch.utils.data import DataLoader

from kaggle_runner.datasets.siim_dataset import SIIMDataset


def provider(
    fold,
    total_folds,
    data_folder,
    df_path,
    phase,
    size,
    mean=None,
    std=None,
    batch_size=8,
    num_workers=2,
):
    df_all = pd.read_csv(df_path)
    df = df_all.drop_duplicates("ImageId")
    df_with_mask = df[df[" EncodedPixels"] != "-1"]
    df_with_mask["has_mask"] = 1
    df_without_mask = df[df[" EncodedPixels"] == "-1"]
    df_without_mask["has_mask"] = 0
    df_without_mask_sampled = df_without_mask.sample(
        len(df_with_mask) + 1500, random_state=2019
    )  # random state is imp
    df = pd.concat([df_with_mask, df_without_mask_sampled])

    # NOTE: equal number of positive and negative cases are chosen.

    kfold = StratifiedKFold(total_folds, shuffle=True, random_state=43)
    train_idx, val_idx = list(kfold.split(df["ImageId"], df["has_mask"]))[fold]
    train_df, val_df = df.iloc[train_idx], df.iloc[val_idx]
    df = train_df if phase == "train" else val_df
    # NOTE: total_folds=5 -> train/val : 80%/20%

    fnames = df["ImageId"].values

    image_dataset = SIIMDataset(df_all, fnames, data_folder, size, mean, std, phase)

    dataloader = DataLoader(
        image_dataset,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=True,
        shuffle=True,
    )
    return dataloader



import errno
import logging
import os
import pickle

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
import torch
import torch.distributed as dist
import torch.utils.data
from IPython.core.debugger import set_trace
from PIL import ImageFile
from tensorflow.python.ops import math_ops

from .utils import logger

trace_flag = True


ImageFile.LOAD_TRUNCATED_IMAGES = True


class AMQPURL:
    class AMQPURL_DEV:
        host = "termite.rmq.cloudamqp.com"  # (Load balanced)
        passwd = "QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI"
        username = "drdsfaew"
        Vhost = "drdsfaew"
        # host = "127.0.0.1"  # (Load balanced)
        # passwd = "guest"
        # username = "guest"
        # Vhost = "/"

    def __init__(
        self,
        host=AMQPURL_DEV.host,
        passwd=AMQPURL_DEV.passwd,
        Vhost=AMQPURL_DEV.Vhost,
        username=AMQPURL_DEV.username,
    ):
        self.host = host
        self.passwd = passwd
        self.Vhost = Vhost
        self.username = username

    def string(self):
        Vhost = self.Vhost

        if self.Vhost == "/":
            Vhost = ""

        return f"amqp://{self.username}:{self.passwd}@{self.host}/{Vhost}"


BIN_FOLDER = (
    "/content/gdrivedata/My Drive/" if os.path.isdir(
        "/content/gdrivedata") else "./"
)

DATASET_BASE_FOLDER = '/kaggle/input'

def dump_obj(obj, filename, fullpath=False, force=False):
    if not fullpath:
        path = BIN_FOLDER + filename
    else:
        path = filename

    if not force and os.path.isfile(path):
        logger.debug(f"{path} already existed, not dumping")
    else:
        logger.debug(f"Overwrite {path}!")
        with open(path, "wb") as f:
            pickle.dump(obj, f)

def get_kaggle_dataset_input(filename):
    filename = os.path.join(DATASET_BASE_FOLDER, filename)

    return get_obj_or_dump(filename, fullpath=True, default=None)

def get_obj_or_dump(filename, fullpath=False, default=None):
    """get_obj_or_dump will dump default obj to file if file not there, otherwise
    obj will be unpickled from file. If file not found, default value will returned."""

    if not fullpath:
        path = BIN_FOLDER + filename
    else:
        path = filename

    if os.path.isfile(path):
        logger.debug("load " + filename)
        with open(path, "rb") as f:
            return pickle.load(f)
    else:
        if default is not None:
            logger.debug("dump :" + filename)
            dump_obj(default, filename)

        return default


def file_exist(filename, fullpath=False):
    if not fullpath:
        path = BIN_FOLDER + filename
    else:
        path = filename

    return os.path.isfile(path)


# 0.5 means no rebalance
def binary_crossentropy_with_focal_seasoned(
    y_true, logit_pred, beta=0.0, gamma=1.0, alpha=0.5, custom_weights_in_Y_true=True
):
    """
    :param alpha:weight for positive classes **loss**. default to 1- true
        positive cnts / all cnts, alpha range [0,1] for class 1 and 1-alpha
        for calss -1.   In practiceαmay be set by inverse class freqency or
        hyperparameter.
    :param custom_weights_in_Y_true:
    :return:
    """
    balanced = gamma * logit_pred + beta
    y_pred = math_ops.sigmoid(balanced)
    # only use gamma in this layer, easier to split out factor

    return binary_crossentropy_with_focal(
        y_true,
        y_pred,
        gamma=0,
        alpha=alpha,
        custom_weights_in_Y_true=custom_weights_in_Y_true,
    )


# 0.5 means no rebalance
def binary_crossentropy_with_focal(
    y_true, y_pred, gamma=1.0, alpha=0.5, custom_weights_in_Y_true=True
):
    """
    :param y_true: y_true
    :param y_pred: y_pred
    :param gamma: make easier ones weights down
    :param alpha: weight for positive classes. default to 1 - (true
        positive cnts / all cnts), alpha range [0,1] for class 1 and 1-alpha
        for calss -1.   In practice α may be set by inverse class freqency or
        hyperparameter.
    :return: bce
    Reference: https://arxiv.org/pdf/1708.02002.pdf

    $$ FL(p_t) = -(1-p_t)^{\gamma}log(p_t) $$
    $$ p_t=p\: if\: y=1$$
    $$ p_t=1-p\: otherwise$$
    """

    # assert 0 <= alpha <= 1 and gamma >= 0
    # hyper parameters, just use the one for binary?
    # alpha = 1. # maybe smaller one can help, as multi-class will make the
    # error larger
    # gamma = 1.5 # for our problem, try different gamma

    # for binary_crossentropy, the implementation is in  tensorflow/tensorflow/python/keras/backend.py
    #       bce = target * alpha* (1-output+epsilon())**gamma * math_ops.log(output + epsilon())
    #       bce += (1 - target) *(1-alpha)* (output+epsilon())**gamma * math_ops.log(1 - output + epsilon())
    # return -bce # binary cross entropy
    eps = tf.keras.backend.epsilon()

    if custom_weights_in_Y_true:
        custom_weights = y_true[:, 1:2]
        y_true = y_true[:, :1]

    if 1.0 - eps <= gamma <= 1.0 + eps:
        bce = alpha * math_ops.multiply(
            1.0 - y_pred, math_ops.multiply(y_true, math_ops.log(y_pred + eps))
        )
        bce += (1 - alpha) * math_ops.multiply(
            y_pred, math_ops.multiply(
                (1.0 - y_true), math_ops.log(1.0 - y_pred + eps))
        )
    elif 0.0 - eps <= gamma <= 0.0 + eps:
        bce = alpha * math_ops.multiply(y_true, math_ops.log(y_pred + eps))
        bce += (1 - alpha) * math_ops.multiply(
            (1.0 - y_true), math_ops.log(1.0 - y_pred + eps)
        )
    else:
        gamma_tensor = tf.broadcast_to(
            tf.constant(gamma), tf.shape(input=y_pred))
        bce = alpha * math_ops.multiply(
            math_ops.pow(1.0 - y_pred, gamma_tensor),
            math_ops.multiply(y_true, math_ops.log(y_pred + eps)),
        )
        bce += (1 - alpha) * math_ops.multiply(
            math_ops.pow(y_pred, gamma_tensor),
            math_ops.multiply(
                (1.0 - y_true), math_ops.log(1.0 - y_pred + eps)),
        )

    if custom_weights_in_Y_true:
        return math_ops.multiply(-bce, custom_weights)
    else:
        return -bce


def reinitLayers(model):
    session = K.get_session()

    for layer in model.layers:
        # if isinstance(layer, keras.engine.topology.Container):

        if isinstance(layer, tf.keras.Model):
            reinitLayers(layer)

            continue
        print("LAYER::", layer.name)

        if layer.trainable is False:
            continue

        for v in layer.__dict__:
            v_arg = getattr(layer, v)

            if hasattr(v_arg, "initializer"):
                # not work for layer wrapper, like Bidirectional
                initializer_method = getattr(v_arg, "initializer")
                initializer_method.run(session=session)
                print("reinitializing layer {}.{}".format(layer.name, v))


# Evaluation metric
# ref https://www.kaggle.com/jesperdramsch/intro-chest-xray-dicom-viz-u-nets-full-data
def dice_coef(y_true, y_pred, smooth=1, threshold=0.5):
    threshold = math_ops.cast(threshold, y_pred.dtype)

    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)

    y_pred_b = math_ops.cast(y_pred_f > threshold, y_pred.dtype)
    y_true_b = math_ops.cast(y_true_f > threshold, y_pred.dtype)

    intersection = K.sum(y_true_b * y_pred_b)

    return (2.0 * intersection + smooth) / (K.sum(y_true_b) + K.sum(y_pred_b) + smooth)


def mask2rle(img, width, height):
    rle = []
    lastColor = 0
    currentPixel = 0
    runStart = -1
    runLength = 0

    for x in range(width):
        for y in range(height):
            currentColor = img[x][y]

            if currentColor != lastColor:
                if currentColor == 255:
                    runStart = currentPixel
                    runLength = 1
                else:
                    rle.append(str(runStart))
                    rle.append(str(runLength))
                    runStart = -1
                    runLength = 0
                    currentPixel = 0
            elif runStart > -1:
                runLength += 1
            lastColor = currentColor
            currentPixel += 1

    return " ".join(rle)


def rle2mask(rle, width, height):
    mask = np.zeros(width * height)
    array = np.asarray([int(x) for x in rle.split()])
    starts = array[0::2]
    lengths = array[1::2]

    current_position = 0

    for index, start in enumerate(starts):
        current_position += start
        mask[current_position: current_position + lengths[index]] = 1
        current_position += lengths[index]

    return mask.reshape(width, height)


def all_gather(data):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors)
    Args:
        data: any picklable object
    Returns:
        list[data]: list of data gathered from each rank
    """
    world_size = get_world_size()

    if world_size == 1:
        return [data]

    # serialized to a Tensor
    buffer = pickle.dumps(data)
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to("cuda")

    # obtain Tensor size of each rank
    local_size = torch.tensor([tensor.numel()], device="cuda")
    size_list = [torch.tensor([0], device="cuda") for _ in range(world_size)]
    dist.all_gather(size_list, local_size)
    size_list = [int(size.item()) for size in size_list]
    max_size = max(size_list)

    # receiving Tensor from all ranks
    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    tensor_list = []

    for _ in size_list:
        tensor_list.append(torch.empty(
            (max_size,), dtype=torch.uint8, device="cuda"))

    if local_size != max_size:
        padding = torch.empty(
            size=(max_size - local_size,), dtype=torch.uint8, device="cuda"
        )
        tensor = torch.cat((tensor, padding), dim=0)
    dist.all_gather(tensor_list, tensor)

    data_list = []

    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    return data_list


def reduce_dict(input_dict, average=True):
    """
    Args:
        input_dict (dict): all the values will be reduced
        average (bool): whether to do average or sum
    Reduce the values in the dictionary from all processes so that all
    processes have the averaged results. Returns a dict with the same fields as
    input_dict, after reduction.
    """
    world_size = get_world_size()

    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # sort the keys so that they are consistent across processes

        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)

        if average:
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}

    return reduced_dict


def collate_fn(batch):
    return tuple(zip(*batch))


def warmup_lr_scheduler(optimizer, warmup_iters, warmup_factor):
    def f(x):
        if x >= warmup_iters:
            return 1
        alpha = float(x) / warmup_iters

        return warmup_factor * (1 - alpha) + alpha

    return torch.optim.lr_scheduler.LambdaLR(optimizer, f)


def mkdir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def setup_for_distributed(is_master):
    """
    This function disables printing when not in master process
    """
    import builtins as __builtin__

    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop("force", False)

        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False

    if not dist.is_initialized():
        return False

    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1

    return dist.get_world_size()


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0

    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def save_on_master(*args, **kwargs):
    if is_main_process():
        torch.save(*args, **kwargs)


def init_distributed_mode(args):
    if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
        args.rank = int(os.environ["RANK"])
        args.world_size = int(os.environ["WORLD_SIZE"])
        args.gpu = int(os.environ["LOCAL_RANK"])
    elif "SLURM_PROCID" in os.environ:
        args.rank = int(os.environ["SLURM_PROCID"])
        args.gpu = args.rank % torch.cuda.device_count()
    else:
        print("Not using distributed mode")
        args.distributed = False

        return

    args.distributed = True

    torch.cuda.set_device(args.gpu)
    args.dist_backend = "nccl"
    print(
        "| distributed init (rank {}): {}".format(args.rank, args.dist_url), flush=True
    )
    torch.distributed.init_process_group(
        backend=args.dist_backend,
        init_method=args.dist_url,
        world_size=args.world_size,
        rank=args.rank,
    )
    torch.distributed.barrier()
    setup_for_distributed(args.rank == 0)


def mask_to_rle(img, width, height):
    rle = []
    lastColor = 0
    currentPixel = 0
    runStart = -1
    runLength = 0

    for x in range(width):
        for y in range(height):
            currentColor = img[x][y]

            if currentColor != lastColor:
                if currentColor == 1:
                    runStart = currentPixel
                    runLength = 1
                else:
                    rle.append(str(runStart))
                    rle.append(str(runLength))
                    runStart = -1
                    runLength = 0
                    currentPixel = 0
            elif runStart > -1:
                runLength += 1
            lastColor = currentColor
            currentPixel += 1

    return " " + " ".join(rle)


def online_mean_and_sd(loader, data_map=None):
    """Compute the mean and sd in an online fashion

        Var[x] = E[X^2] - E^2[X]
    credit xwkuang5
    @https://discuss.pytorch.org/t/about-normalization-using-pre-trained-vgg16-networks/23560/7
    """
    cnt = 0
    # fst_moment = torch.empty(3)
    # snd_moment = torch.empty(3)
    fst_moment = np.zeros(3)
    snd_moment = np.zeros(3)

    for data in loader:
        if data_map is not None:
            data = data_map(data)
        data = np.array([t.numpy() for t in data])
        b, c, h, w = data.shape  # data here is tuple... if loader batch > 1
        nb_pixels = b * h * w
        # sum_ = torch.sum(data, dim=[0, 2, 3])
        sum_ = data.sum(axis=0).sum(axis=-1).sum(axis=-1)
        # sum_of_square = torch.sum(data ** 2, dim=[0, 2, 3])
        sum_of_square = (data ** 2).sum(axis=0).sum(axis=-1).sum(axis=-1)
        fst_moment = (cnt * fst_moment + sum_) / (cnt + nb_pixels)
        snd_moment = (cnt * snd_moment + sum_of_square) / (cnt + nb_pixels)

        cnt += nb_pixels

    return fst_moment, np.sqrt(snd_moment - fst_moment ** 2)


def download_file_one_at_a_time(file_name, directory=".", overwrite=True):
    if overwrite:
        run_process_print(
            'wget http://23.105.212.181:8000/{0} -O "{1}/{0}"'.format(
                file_name, directory
            )
        )
    else:
        run_process_print(
            "[ -f {1}/{0} ] || wget http://23.105.212.181:8000/{0} -P {1}".format(
                file_name, directory
            )
        )


def run_process_print(command_str):
    print(command_str.split())
    get_ipython().system(command_str)



import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
import torch

from kaggle_runner import may_debug


def matthews_correlation_aux_stripper(y_true, y_pred):
    ts = y_true.shape
    ps = y_pred.shape

    if len(ps) > len(ts) or (
            (ps[1] > ts[1]) if ts[1] is not None else False):
        y_pred = y_pred[:,0]

    return matthews_correlation(y_true, y_pred)

def matthews_correlation(y_true, y_pred):
    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    y_pred_neg = 1 - y_pred_pos

    y_pos = K.round(K.clip(y_true, 0, 1))
    y_neg = 1 - y_pos

    tp = K.sum(y_pos * y_pred_pos)
    tn = K.sum(y_neg * y_pred_neg)

    fp = K.sum(y_neg * y_pred_pos)
    fn = K.sum(y_pos * y_pred_neg)

    numerator = (tp * tn - fp * fn)
    denominator = K.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))

    return numerator / (denominator + K.epsilon())


def metric(probability, truth, threshold=0.5, reduction="none"):
    """Calculates dice of positive and negative images seperately"""
    """probability and truth must be torch tensors"""
    batch_size = len(truth)
    with torch.no_grad():
        probability = probability.view(batch_size, -1)
        truth = truth.view(batch_size, -1)
        assert probability.shape == truth.shape

        p = (probability > threshold).float()
        t = (truth > 0.5).float()

        t_sum = t.sum(-1)
        p_sum = p.sum(-1)
        neg_index = torch.nonzero(t_sum == 0)
        pos_index = torch.nonzero(t_sum >= 1)

        dice_neg = (p_sum == 0).float()
        dice_pos = 2 * (p * t).sum(-1) / ((p + t).sum(-1))

        dice_neg = dice_neg[neg_index]
        dice_pos = dice_pos[pos_index]
        dice = torch.cat([dice_pos, dice_neg])

        dice_neg = np.nan_to_num(dice_neg.mean().item(), 0)
        dice_pos = np.nan_to_num(dice_pos.mean().item(), 0)
        dice = dice.mean().item()

        num_neg = len(neg_index)
        num_pos = len(pos_index)

    return dice, dice_neg, dice_pos, num_neg, num_pos


def binary_sensitivity_np(y_pred, y_true):
    threshold = 0.5
    # predict_false = y_pred <= threshold
    y_true = y_true > threshold
    predict_true = y_pred > threshold
    TP = np.multiply(y_true, predict_true)
    # FP = np.logical_and(y_true == 0, predict_true)

    # as Keras Tensors
    TP = TP.sum()
    # FP = FP.sum()

    sensitivity = TP / y_true.sum()

    return sensitivity


def binary_sensitivity(y_pred, y_true):
    """Compute the confusion matrix for a set of predictions.

    Parameters
    ----------
    y_pred   : predicted values for a batch if samples (must be binary: 0 or 1)
    y_true   : correct values for the set of samples used (must be binary: 0 or 1)

    Returns
    -------
    out : the specificity
    """
    threshold = 0.5
    TP = np.logical_and(K.eval(y_true) == 1, K.eval(y_pred) <= threshold)
    FP = np.logical_and(K.eval(y_true) == 0, K.eval(y_pred) > threshold)

    # as Keras Tensors
    TP = K.sum(K.variable(TP))
    FP = K.sum(K.variable(FP))

    sensitivity = TP / (TP + FP + K.epsilon())

    return sensitivity


def binary_specificity(y_pred, y_true):
    """Compute the confusion matrix for a set of predictions.

    Parameters
    ----------
    y_pred   : predicted values for a batch if samples (must be binary: 0 or 1)
    y_true   : correct values for the set of samples used (must be binary: 0 or 1)

    Returns
    -------
    out : the specificity
    """

    threshold = 0.5
    TN = np.logical_and(K.eval(y_true) == 0, K.eval(y_pred) <= threshold)
    FP = np.logical_and(K.eval(y_true) == 0, K.eval(y_pred) > threshold)

    # as Keras Tensors
    TN = K.sum(K.variable(TN))
    FP = K.sum(K.variable(FP))

    specificity = TN / (TN + FP + K.epsilon())

    return specificity


def binary_auc_probability(y_true, y_pred, threshold=0.5, N_MORE=True, epsilon=1e-12):
    """
    refer to this: https://blog.revolutionanalytics.com/2016/11/calculating-auc.html

    The probabilistic interpretation is that if you randomly choose a positive case and a negative case, the probability that the positive case outranks the negative case according to the classifier is given by the AUC. This is evident from the figure, where the total area of the plot is normalized to one, the cells of the matrix enumerate all possible combinations of positive and negative cases, and the fraction under the curve comprises the cells where the positive case outranks the negative one.

    :param y_true:
    :param y_pred:
    :param threshold:
    :return: accuracy, f1 for this batch... not the global one, we need to be careful!!
    """

    # labels: y_true, scores: y_pred, N the size of sample
    # auc_probability < - function(labels, scores, N=1e7)
    # {
    #     pos < - sample(scores[labels], N, replace=TRUE)
    # neg < - sample(scores[!labels], N, replace = TRUE)
    # # sum( (1 + sign(pos - neg))/2)/N # does the same thing
    # (sum(pos > neg) + sum(pos == neg) / 2) / N  # give partial credit for ties
    # }

    # auc_probability(as.logical(category), prediction)

    threshold = math_ops.cast(threshold, y_pred.dtype)
    # y_pred = math_ops.cast(y_pred > threshold, y_pred.dtype)
    y_true = math_ops.cast(y_true > threshold, y_pred.dtype)

    true_pos_predict = math_ops.multiply(y_true, y_pred)  # %6 pos
    # 94% neg...
    true_neg_predict = math_ops.multiply(1.0 - y_true, 1 - y_pred)

    # recision = math_ops.div(correct_pos, predict_pos)
    # recall = math_ops.div(correct_pos, ground_pos)

    # if N_MORE:
    #    m = (2*recall*precision) / (precision+recall)
    # else:
    #    #m = (sensitivity + specificity)/2 # balanced accuracy
    #    raise NotImplementedError("Balanced accuracy metric is not implemented")

    # return m

def bin_prd_clsf_info_neg(y_true, y_pred, threshold=0.5, N_MORE=True, epsilon=1e-7):
    """
    refer to this: https://stats.stackexchange.com/questions/49579/balanced-accuracy-vs-f-1-score

    Both F1 and b_acc are metrics for classifier evaluation, that (to some extent) handle class imbalance. Depending
     of which of the two classes (N or P) outnumbers the other, each metric is outperforms the other.

    1) If N >> P, f1 is a better.

    2) If P >> N, b_acc is better.

    For code: refer to this: https://www.kaggle.com/c/quora-insincere-questions-classification/discussion/70841

    :param y_true:
    :param y_pred:
    :param threshold:
    :return: accuracy, f1 for this batch... not the global one, we need to be careful!!
    """
    # if FOCAL_LOSS_GAMMA == 2.0:
    #    threshold = 0.57
    # elif FOCAL_LOSS_GAMMA == 1.0:
    #    threshold = (0.53 + (
    #                0.722 - 0.097)) / 2  # (by...reading the test result..., found it changes every training... so useless)
    threshold = math_ops.cast(threshold, y_pred.dtype)
    y_pred_b = math_ops.cast(y_pred > threshold, y_pred.dtype)
    y_true_b = math_ops.cast(y_true > threshold, y_pred.dtype)

    # ground_pos = math_ops.reduce_sum(y_true) + epsilon
    # correct_pos = math_ops.reduce_sum(math_ops.multiply(y_true, y_pred)) + epsilon
    # predict_pos = math_ops.reduce_sum(y_pred) + epsilon
    true_cnt = math_ops.reduce_sum(y_true_b) + epsilon
    false_cnt = math_ops.reduce_sum(1 - y_true_b) + epsilon
    pred_true_cnt = math_ops.reduce_sum(y_pred_b) + epsilon
    pred_false_cnt = math_ops.reduce_sum(1 - y_pred_b) + epsilon
    # true_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, y_true_b))/true_cnt
    # false_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, 1-y_true_b))/false_cnt

    # true_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, y_true_b))/true_cnt
    # false_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, 1-y_true_b))/false_cnt

    # tp_mean_scaled = math_ops.cast(true_predict_mean*100, tf.int8)
    # tp_mean_scaled = math_ops.cast(tp_mean_scaled, tf.float32)
    # precision = math_ops.div(correct_pos, predict_pos)
    # recall = math_ops.div(correct_pos, ground_pos)

    # if N_MORE:
    #    m = (2 * recall * precision) / (precision + recall)
    # else:
    #    # m = (sensitivity + specificity)/2 # balanced accuracy
    #    raise NotImplementedError("Balanced accuracy metric is not implemented")

    return (pred_false_cnt - false_cnt) / false_cnt  # (batchsize 1024)

def bin_prd_clsf_info_pos(y_true, y_pred, threshold=0.5, N_MORE=True, epsilon=1e-7):
    """
    refer to this: https://stats.stackexchange.com/questions/49579/balanced-accuracy-vs-f-1-score

    Both F1 and b_acc are metrics for classifier evaluation, that (to some extent) handle class imbalance. Depending
     of which of the two classes (N or P) outnumbers the other, each metric is outperforms the other.

    1) If N >> P, f1 is a better.

    2) If P >> N, b_acc is better.

    For code: refer to this: https://www.kaggle.com/c/quora-insincere-questions-classification/discussion/70841

    :param y_true:
    :param y_pred:
    :param threshold:
    :return: accuracy, f1 for this batch... not the global one, we need to be careful!!
    """
    # if FOCAL_LOSS_GAMMA == 2.0:
    #    threshold = 0.57
    # elif FOCAL_LOSS_GAMMA == 1.0:
    #    threshold = (0.53 + (
    #                0.722 - 0.097)) / 2  # (by...reading the test result..., found it changes every training... so useless)
    threshold = math_ops.cast(threshold, y_pred.dtype)
    y_pred_b = math_ops.cast(y_pred > threshold, y_pred.dtype)
    y_true_b = math_ops.cast(y_true > threshold, y_pred.dtype)

    # ground_pos = math_ops.reduce_sum(y_true) + epsilon
    # correct_pos = math_ops.reduce_sum(math_ops.multiply(y_true, y_pred)) + epsilon
    # predict_pos = math_ops.reduce_sum(y_pred) + epsilon
    true_cnt = math_ops.reduce_sum(y_true_b) + epsilon
    false_cnt = math_ops.reduce_sum(1 - y_true_b) + epsilon
    pred_true_cnt = math_ops.reduce_sum(y_pred_b) + epsilon
    pred_false_cnt = math_ops.reduce_sum(1 - y_pred_b) + epsilon
    # true_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, y_true_b))/true_cnt
    # false_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, 1-y_true_b))/false_cnt

    # true_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, y_true_b))/true_cnt
    # false_predict_mean = math_ops.reduce_sum(math_ops.multiply(y_pred, 1-y_true_b))/false_cnt

    # tp_mean_scaled = math_ops.cast(true_predict_mean*100, tf.int8)
    # tp_mean_scaled = math_ops.cast(tp_mean_scaled, tf.float32)
    # precision = math_ops.div(correct_pos, predict_pos)
    # recall = math_ops.div(correct_pos, ground_pos)

    # if N_MORE:
    #    m = (2 * recall * precision) / (precision + recall)
    # else:
    #    # m = (sensitivity + specificity)/2 # balanced accuracy
    #    raise NotImplementedError("Balanced accuracy metric is not implemented")

    return (pred_true_cnt - true_cnt) / true_cnt  # (batchsize 1024)



from unittest import TestCase

from kaggle_runner.logs import NBatchProgBarLogger


class TestLogs(TestCase):
    @classmethod
    def setup_class(cls):
        "prepare a dummy model for logger to show"
        # TODO set up a dummy model / or just use imagenette
        cls.model = None

    @classmethod
    def teardown_class(cls):
        print("teardown_class called once for the class")

    def test_NBatchProgBarLogger(self):
        assert self.model is not None
        self.BS = 100

        self.model.fit(
            [X_train_aux, X_train_main],
            Y_train,
            batch_size=BS,
            verbose=0,
            callbacks=[NBatchProgBarLogger()],
        )



import torch
import torch.nn as nn
from transformers import XLMRobertaModel, DistilBertModel
from transformers import XLNetConfig, XLNetModel, XLNetPreTrainedModel

BACKBONE_PATH = 'xlm-roberta-large'


class ToxicSimpleNNModel(nn.Module):
    def __init__(self, use_aux=True):
        super(ToxicSimpleNNModel, self).__init__()
        self.backbone = XLMRobertaModel.from_pretrained(BACKBONE_PATH)
        self.dropout = nn.Dropout(0.3)
        aux_len = 0

        if use_aux:
            aux_len = 5
        self.linear = nn.Linear(
            in_features=self.backbone.pooler.dense.out_features*2,
            out_features=2+aux_len,
        )

    def forward(self, input_ids, attention_masks):
        bs, seq_length = input_ids.shape
        seq_x, _ = self.backbone(
            input_ids=input_ids, attention_mask=attention_masks)
        apool = torch.mean(seq_x, 1)
        mpool, _ = torch.max(seq_x, 1)
        x = torch.cat((apool, mpool), 1)
        x = self.dropout(x)

        return self.linear(x)



from kaggle_runner.kernels.kernel import KaggleKernelOnlyPredict


from transformers import *

class ToxicPredictModel(KaggleKernelOnlyPredict):
    def __init__(self, model_path):
        super(ToxicPredictModel, self).__init__(model_path)
        self.only_predict = True
        self.model_path = model_path

    def save_model(self, output_dir="./models/"):
        model = self.model

        from transformers import WEIGHTS_NAME, CONFIG_NAME
# Step 1: Save a model, configuration and vocabulary that you have fine-tuned

# If we have a distributed model, save only the encapsulated model
# (it was wrapped in PyTorch DistributedDataParallel or DataParallel)
        model_to_save = model.module if hasattr(model, 'module') else model

# If we save using the predefined names, we can load using `from_pretrained`
        output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
        output_config_file = os.path.join(output_dir, CONFIG_NAME)

        torch.save(model_to_save.state_dict(), output_model_file)
        model_to_save.config.to_json_file(output_config_file)
        tokenizer.save_pretrained(output_dir)

    def build_and_set_model(self):
        """load pretrained one"""
        # Example for a Bert model
        #config = BertConfig.from_json_file(output_config_file)
        #model = BertForQuestionAnswering(config)
        #state_dict = torch.load(output_model_file)
        #model.load_state_dict(state_dict)
        #tokenizer = BertTokenizer(output_vocab_file, do_lower_case=args.do_lower_case)
        model = BertForQuestionAnswering.from_pretrained(output_dir)
        tokenizer = BertTokenizer.from_pretrained(output_dir)  # Add specific options if needed

    def prepare_train_dev_data(self):
        pass

    def prepare_test_data(self, data_config=None):
        pass

    def check_predict_details(self):
        pass

    def peek_data(self):
        pass

def only_predict():
    pass

def test_init():
    k = ToxicPredictModel(".")
    assert k is not None
    k.save_model()

import torch

def test_load():
    output_model_file='/kaggle/input/bert-for-toxic-classfication-trained/2020-06-21_XLMRobertaModel_tpu_trained.bin'
    state_dict = torch.load(output_model_file)
    k.model.load_state_dict(state_dict)

    print(model)



import json
import logging
import types
from socket import gethostname

import parse
from python_logging_rabbitmq import RabbitMQHandler
from python_logging_rabbitmq.compat import text_type


class AMQPURL:
    class __AMQPURL_DEV:
        #host = "termite.rmq.cloudamqp.com"  # (Load balanced)
        #passwd = "QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI"  # oh~ just give my password out~
        #username = "drdsfaew"
        #Vhost = "drdsfaew"
        host = "pengyuzhou.com"
        passwd = "9b83ca70cf4cda89524d2283a4d675f6"
        username = "kaggle"
        port = "5672"
        Vhost = "/"

    def __init__(
        self,
        host=__AMQPURL_DEV.host,
        passwd=__AMQPURL_DEV.passwd,
        Vhost=__AMQPURL_DEV.Vhost,
        username=__AMQPURL_DEV.username,
        port=__AMQPURL_DEV.port
    ):
        self.host = host
        self.passwd = passwd
        self.Vhost = Vhost
        self.username = username
        self.port = port

    def string(self):
        Vhost = self.Vhost

        if self.Vhost == "/":
            Vhost = ""

        return f"amqp://{self.username}:{self.passwd}@{self.host}/{Vhost}"


def get_logger(name="utils", level=logging.DEBUG):
    "get_logger just return basic logger, no AMQP included"
    FORMAT = "[%(levelname)s]%(asctime)s:%(name)s:%(message)s"
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger(name)
    logger.setLevel(level)

    return logger

def log_format(self, record):
    data = record.__dict__.copy()

    if record.args:
        msg = record.msg % record.args
    else:
        msg = record.msg

    data.update(
        host=gethostname(), msg=msg, args=tuple(text_type(arg) for arg in record.args)
    )

    if "exc_info" in data and data["exc_info"]:
        data["exc_info"] = self.formatException(data["exc_info"])

    if self.include:
        data = {f: data[f] for f in self.include}
    elif self.exclude:
        for f in self.exclude:
            if f in data:
                del data[f]

    return json.dumps(data)

def attach_data_collector(logger, AMQPURL=AMQPURL()):
    """
    Credits: https: // github.com/albertomr86/python-logging-rabbitmq

    !pip install python_logging_rabbitmq

    """

    rabbit = RabbitMQHandler(
        host=AMQPURL.host,
        port=AMQPURL.port,
        username=AMQPURL.username,
        password=AMQPURL.passwd,
        exchange="logs_topic",
        connection_params={"virtual_host": AMQPURL.Vhost},
        declare_exchange=True,
    )
    # rabbit.connection_params["virtual_host"] = self.AMQPURL.Vhost create
    # kernel and run
    rabbit.formatter.format = types.MethodType(
        log_format, rabbit.formatter)
    logger.addHandler(rabbit)

    return logger

logger = attach_data_collector(get_logger())

def parse_AMQP(url_str):
    Vhost = None
    res = parse.parse("amqp://{}:{}@{}/{}", url_str)

    if res is None:
        res = parse.parse("amqp://{}:{}@{}/", url_str)

        if res is None:
            raise RuntimeError("AMQP URL error")
        else:
            Vhost = "/"
            username, passwd, host = res
    else:
        username, passwd, host, Vhost = res

    try:
        return AMQPURL(host, passwd, Vhost, username)
    except TypeError as e:
        logger.debug(e)



import os
from pathlib import Path
import pandas as pd
import torch
from torch import nn
from fastai import *
from fastai import vision
from fastai.basic_data import *
from fastai.basic_train import Learner
from fastai.callbacks import CSVLogger
from fastai.core import *
from fastai.torch_core import *
from fastai.vision import *

from .kernel import KaggleKernel
from .. import logger, may_debug


class FastAIKernel(KaggleKernel):
    """fast.ai thing
    """
    def build_and_set_model(self):
        self.model = None

    def check_predict_details(self):
        assert False

    def peek_data(self):
        for targets, tokens, attention_masks in self.train_dataset:
            break

        print(targets)
        print(tokens.shape)
        print(attention_masks.shape)

        for targets, tokens, attention_masks in self.validation_dataset:
            break

        print(targets)
        print(tokens.shape)
        print(attention_masks.shape)

        for ids, tokens, attention_masks in self.test_dataset:
            break

        print(ids)
        print(tokens.shape)
        print(attention_masks.shape)

    def set_loss(self, loss_func):
        self.model_loss = loss_func

    def set_metrics(self, metrics):
        self.model_metrics = metrics

    def __init__(self, **kargs):
        super(FastAIKernel, self).__init__(logger=logger)
        self.developing = True
        self.learner = None

        for required in ['opt_func', 'loss_func', 'metrics']:
            assert required in kargs

        def _map(name):
            name_map = {'metrics': 'model_metrics'}

            return name_map[name] if name in name_map else name

        for k, v in kargs.items():
            setattr(self, _map(k), v)

    @classmethod
    def create_learner(cls, k=None, data=None, model=None, opt_func=None, loss_func=None, metrics=None, **kargs):
        """opt_func should pass seprately"""

        if k is not None:
            data = k.data if hasattr(
                k, 'data') and k.data is not None else data
            model = k.model if hasattr(
                k, 'model') and k.model is not None else model
            loss_func = k.model_loss if hasattr(
                k, 'model_loss') and k.model_loss is not None else loss_func
            metrics = k.model_metrics if hasattr(
                k, 'model_metrics') and k.model_metrics is not None else metrics

        assert data is not None
        assert opt_func is not None
        learner = Learner(data, model, opt_func, loss_func=loss_func, metrics=metrics, bn_wd=False, **kargs) # opt_func postitional parameter is before loss_func

        return learner


def test_learner_init():
    l = FastAIKernel(loss_func=None, metrics=None)
    assert l is not None


def test_learner_fit():
    path = untar_data(URLs.MNIST_SAMPLE)
    data = ImageDataBunch.from_folder(path)

    #INPUT = Path("../input/digit-recognizer")
    #os.listdir(INPUT)

    #train_df = pd.read_csv(INPUT/"train.csv")

    #test_df = pd.read_csv(INPUT/"test.csv")

    #TRAIN = Path("../train")
    #TEST = Path("../test")

    #for index in range(10):
    #    try:
    #        os.makedirs(TRAIN/str(index))
    #    except:
    #        pass

    #sorted(os.listdir(TRAIN))

    #try:
    #    os.makedirs(TEST)
    #except:
    #    pass

    #from PIL import Image

    #def saveDigit(digit, filepath):
    #    digit = digit.reshape(28, 28)
    #    digit = digit.astype(np.uint8)

    #    img = Image.fromarray(digit)
    #    img.save(filepath)

    ##for index, row in train_df.iterrows():

    ##    label,digit = row[0], row[1:]

    ##    folder = TRAIN/str(label)
    ##    filename = f"{index}.jpg"
    ##    filepath = folder/filename

    ##    digit = digit.values

    ##    saveDigit(digit, filepath)

    ##for index, digit in test_df.iterrows():

    ##    folder = TEST
    ##    filename = f"{index}.jpg"
    ##    filepath = folder/filename

    ##    digit = digit.values

    ##    saveDigit(digit, filepath)

    #import matplotlib.pyplot as plt

    #def displayTrainingData():
    #    fig = plt.figure(figsize=(5, 10))

    #    for rowIndex in range(1, 10):
    #        subdirectory = str(rowIndex)
    #        path = TRAIN/subdirectory
    #        images = os.listdir(path)

    #        for sampleIndex in range(1, 6):
    #            randomNumber = random.randint(0, len(images)-1)
    #            image = Image.open(path/images[randomNumber])
    #            ax = fig.add_subplot(10, 5, 5*rowIndex + sampleIndex)
    #            ax.axis("off")

    #            plt.imshow(image, cmap='gray')

    #    plt.show()

    #def displayTestingData():
    #    fig = plt.figure(figsize=(5, 10))

    #    paths = os.listdir(TEST)

    #    for i in range(1, 51):
    #        randomNumber = random.randint(0, len(paths)-1)
    #        image = Image.open(TEST/paths[randomNumber])

    #        ax = fig.add_subplot(10, 5, i)
    #        ax.axis("off")

    #        plt.imshow(image, cmap='gray')
    #    plt.show()

    ##print('samples of training data')
    ##displayTrainingData()

    ##print('samples of testing data')
    ##displayTestingData()

    ##image_path = TEST/os.listdir(TEST)[9]
    ##image = Image.open(image_path)
    ##image_array = np.asarray(image)

#fig#, ax = plt.subplots(figsize=(15, 15))

#img# = ax.imshow(image_array, cmap='gray')

#for# x in range(28):
#   # for y in range(28):
#   #     value = round(image_array[y][x]/255.0, 2)
#   #     color = 'black' if value > 0.5 else 'white'
#   #     ax.annotate(s=value, xy=(x, y), ha='center', va='center', color=color)
#
#plt#.axis('off')
#plt#.show()

    #tfms = get_transforms(do_flip=False)

    #data = ImageDataBunch.from_folder(
    #    path=str(TRAIN),
    #    test=str(TEST),
    #    valid_pct=0.2,
    #    bs=16,
    #    size=28,
    #    #num_workers = 0,
    #    ds_tfms=tfms
    #)
    ##print(mnist_stats)
    #data.normalize(mnist_stats)
    ##print(data.classes)

    ##learn = cnn_learner(data, base_arch=models.resnet18, metrics=accuracy,
    ##                    model_dir="/tmp/models", callback_fns=ShowGraph)
    ##learn.fit_one_cycle(cyc_len=5)
    ##interp = ClassificationInterpretation.from_learner(learn)
    ##interp.plot_top_losses(9, figsize=(7, 7))
    ##interp.plot_confusion_matrix()
    ##import platform; platform.system()

    ##flip_tfm = RandTransform(tfm=TfmPixel (flip_lr), kwargs={}, p=1, resolved={}, do_run=True, is_random=True, use_on_y=True)
    ##folder = TRAIN/"3"
    ##filename = os.listdir(folder)[0]
    ##img = open_image(TRAIN/folder/filename)
    ##display(img)
    ##display(img.apply_tfms(flip_tfm))

    ##tfms = get_transforms(do_flip=False)

    ##learn = cnn_learner(data, base_arch=models.densenet169, metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)
    ###learn = cnn_learner(data, base_arch=models.densenet169, pretrained=False, metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)
    ###import torchvision.models
    ###learn = Learner(data, torchvision.models.googlenet(pretrained=True), metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)
    ##batch_size = 16

    class CNN(nn.Module):
        def __init__(self):
            super(CNN, self).__init__()
            # input is 28 pixels x 28 pixels x 3 channels
            # our original data was grayscale, so only one channel, but fast.ai automatically loads in the data as RGB
            self.conv1 = nn.Conv2d(3, 16, 3, padding=1)
            self.conv2 = nn.Conv2d(16, 32, 3, padding=1)
            self.pool = nn.MaxPool2d(2, 2)
            self.fc1 = nn.Linear(7*7*32, 500)
            self.fc2 = nn.Linear(500, 10)
            self.relu = nn.ReLU()

        def forward(self, x):
            # print(x.size())
            # x (28x28x3)
            x = self.conv1(x)
            # x (28x28x16)
            x = self.pool(x)
            # x (14x14x16)
            x = self.relu(x)

            x = self.conv2(x)
            # x (14x14x32)
            x = self.pool(x)
            # x (7x7x32)
            x = self.relu(x)

            # flatten images in batch
            #print(x.size())
            x = x.view(-1,7*7*32)
            #print(x.size())
            x = self.fc1(x)
            x = self.relu(x)

            x = self.fc2(x)
            x = self.relu(x)

            return x

    ##learn = Learner(data, CNN(), metrics=accuracy, model_dir="/tmp/models", callback_fns=ShowGraph)

    k = FastAIKernel(loss_func=None, metrics=accuracy) # loss_func will be given a default one
    #k.setup_learner(data=data, model=models.resnet18, metrics=None)
    k.create_learner(data=data, model=CNN(), metrics=None)
    assert k.learner

    k.learner.lr_find()
    k.learner.recorder.plot()
    k.learner.fit_one_cycle(1) # just test basic function

def test_basic_model():
    m = nn.Dropout(p=0.2)
    input = torch.randn(20, 16)
    output = m(input)
    assert sum(sum(input.numpy())) != sum(sum(output.numpy()))
